"""
Here we calculate weighted essentially non-oscillatory 
(WENO) fluxes for the MHD equations.

The idea of WENO is to find interface fluxes by interpolating
the cell centered fluxes using several stencils, and then
weighting the stencils based on their smoothness.

The reconstruction is done in characteristic variables to
better capture the underlying wave structure. At each interface,
we compute the eigenstructure (evaluated at the average of the
left and right states), and project all stencil
characteristic space.

Consider the interface at i + 1/2. Our vector of conserved
variables is q = (rho, rho*v_x, rho*v_y, rho*v_z, B_x, B_y, B_z, E)^T
with N_vars = 8 variables. In the eigenstructure of the MHD equations,
we have N_char = 7 characteristic waves.

We calculate the flux as follows:

1. We retrieve the eigenstructure given by the right 
   and left eigenvector matrices R_{i+1/2} \in R^{N_vars x N_char} 
   and L_{i+1/2} \in R^{N_char x N_vars}, as well 
   as the eigenvalues lambda at 
   q_{i+1/2} ~ 0.5 * (q_i + q_{i+1}).

2. In the stencil m = i - 2, ..., i + 2, we project the fluxes
   F_m and conserved variables q_m into characteristic space:
   F_s_m = L^s_{i+1/2} * F_m, q_s_m = L^s_{i+1/2} * q_m, where L^s_{i+1/2}
   is the s-th row of L so F_s_m and q_s_m are scalar fields. All
   fluxes and conserved variables in the stencil m = i - 2, ..., i + 2
   are projected using the same L^s_{i+1/2} at the interface i + 1/2.

3. We compute the differences ΔF_s_{m+1/2} = F_s_{m+1} - F_s_m and
   Δq_s_{m+1/2} = q_s_{m+1} - q_s_m for m = i - 2, ..., i + 1.

4. We use local Lax-Friedrichs flux splitting to split the fluxes
   into F_s^+ and F_s^- such that \partial_u F_s^+ only has non-negative
   eigenvalues, and \partial_u F_s^- only has non-positive eigenvalues.
   Both can then be properly upwinded with skewed stencils (for F_s^+ we
   use a left-biased stencil, for F_s^- we use a right-biased stencil, 
   see step 5).

   ΔF_s^+_{m+1/2} = 0.5 * (ΔF_s_{m+1/2} + alpha^s * Δq_s_{m+1/2}), m = i - 2, ..., i + 1
   ΔF_s^-_{m+1/2} = 0.5 * (ΔF_s_{m+1/2} - alpha^s * Δq_s_{m+1/2}), m = i - 1, ..., i + 2

   where alpha^s = max(|lambda^s_m|) over the 
   stencil m = i - 2, ..., i + 3.

5. We can compactly write the WENO flux reconstruction as:
   
   F_{i+1/2} = 1/12 * (-F_{i-1} + 7*F_i + 7*F_{i+1} - F_{i+2})
                +sum_{s = 1}^{N_char} [
                    -\phi(ΔF_s^+_{i-3/2}, ΔF_s^+_{i-1/2}, ΔF_s^+_{i+1/2}, ΔF_s^+_{i+3/2})
                    +\phi(ΔF_s^-_{i+5/2}, ΔF_s^-_{i+3/2}, ΔF_s^-_{i+1/2}, ΔF_s^-_{i-1/2})
                ] * R^s_{i+1/2}
    
    where R^s_{i+1/2} is the s-th column of R at the interface i + 1/2,
    and \phi is the WENO interpolant function given by:

    \phi(a, b, c, d) = 1/3 ω_0 (a - 2b + c) + 1/6 (ω_2 - 1/2) (b - 2c + d)

    with weight functions:
    
    ω_0 = α_0 / (α_0 + α_1 + α_2)
    ω_2 = α_2 / (α_0 + α_1 + α_2)

    α_0 = 1 / (ε + IS_0)^2
    α_1 = 6 / (ε + IS_1)^2
    α_2 = 3 / (ε + IS_2)^2

    and smoothness indicators:

    IS_0 = 13 (a - b)^2 + 3 (a - 3b)^2
    IS_1 = 13 (b - c)^2 + 3 (b + c)^2
    IS_2 = 13 (c - d)^2 + 3 (3c - d)^2

    ε is a small parameter to avoid 
    division by zero, here taken as 1e-8.

NOTE: I have seen formulations where the first part of the flux
is also calculated in characteristic space and then transformed back,
but I found that at single precision this introduces small perturbations
as RL is not exactly the identity matrix by finite precision effects.

For literature references, see:

 - High Order ENO and WENO Schemes for Computational Fluid Dynamics by Chi-Wang Shu (1997)
   (https://doi.org/10.1007/978-3-662-03882-6_5)

Concretely we implement the 5th-order WENO scheme as described in 

- HOW-MHD: A High-Order WENO-Based Magnetohydrodynamic Code with a High-Order 
  Constrained Transport Algorithm for Astrophysical Applications by Seo & Ryu 2023
  (https://arxiv.org/abs/2304.04360)
"""

from functools import partial
import jax
import jax.numpy as jnp
from typing import Union

from jax import checkpoint

from astronomix._finite_difference._fluid_equations._eigen import _eigen_L_row, _eigen_R_col, _eigen_lambdas
from astronomix._finite_difference._fluid_equations._fluxes import _mhd_flux_x
from astronomix.option_classes.simulation_config import SimulationConfig
from astronomix.variable_registry.registered_variables import RegisteredVariables

@partial(jax.jit, static_argnames=["registered_variables", "config"])
def _weno_flux_x(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
):
    """
    WENO flux reconstruction.
    """

    epsilon = 1e-8

    # retrieve the center fluxes
    F = _mhd_flux_x(conserved_state, rhomin, pgmin, gamma, config, registered_variables)
    
    # with this we can already compute the first part of the flux
    F_interface = 1/12 * (
        -jnp.roll(F, 1, axis=1) + 7 * F + 7 * jnp.roll(F, -1, axis=1) - jnp.roll(F, -2, axis=1)
    )

    def mode_flux(mode, F_current):

        # get eigenstructure for this mode
        lambdas_center = _eigen_lambdas(conserved_state, rhomin, pgmin, gamma, registered_variables, mode)
        L_row = _eigen_L_row(conserved_state, rhomin, pgmin, gamma, registered_variables, mode)

        F0 = jnp.roll(F,  2, axis=1)   # shape (N_vars, Nx, Ny, Nz) — i-2 at target i
        F1 = jnp.roll(F,  1, axis=1)   # i-1
        F2 = F                         # i
        F3 = jnp.roll(F, -1, axis=1)   # i+1
        F4 = jnp.roll(F, -2, axis=1)   # i+2
        F5 = jnp.roll(F, -3, axis=1)   # i+3

        s0 = jnp.einsum('nxyz,nxyz->xyz', L_row, F0)
        s1 = jnp.einsum('nxyz,nxyz->xyz', L_row, F1)
        s2 = jnp.einsum('nxyz,nxyz->xyz', L_row, F2)
        s3 = jnp.einsum('nxyz,nxyz->xyz', L_row, F3)
        s4 = jnp.einsum('nxyz,nxyz->xyz', L_row, F4)
        s5 = jnp.einsum('nxyz,nxyz->xyz', L_row, F5)

        q0 = jnp.einsum('nxyz,nxyz->xyz', L_row, jnp.roll(conserved_state, 2, axis=1))
        q1 = jnp.einsum('nxyz,nxyz->xyz', L_row, jnp.roll(conserved_state, 1, axis=1))
        q2 = jnp.einsum('nxyz,nxyz->xyz', L_row, conserved_state)
        q3 = jnp.einsum('nxyz,nxyz->xyz', L_row, jnp.roll(conserved_state, -1, axis=1))
        q4 = jnp.einsum('nxyz,nxyz->xyz', L_row, jnp.roll(conserved_state, -2, axis=1))
        q5 = jnp.einsum('nxyz,nxyz->xyz', L_row, jnp.roll(conserved_state, -3, axis=1))

        # dFsk identical to original: d0 = s1 - s0, d1 = s2 - s1, ...
        d0 = s1 - s0
        d1 = s2 - s1
        d2 = s3 - s2
        d3 = s4 - s3
        d4 = s5 - s4

        dq0 = q1 - q0
        dq1 = q2 - q1
        dq2 = q3 - q2
        dq3 = q4 - q3
        dq4 = q5 - q4

        # compute amx over the same stencil (take abs then max over the six entries)
        lam0 = jnp.roll(lambdas_center,  2, axis=0)
        lam1 = jnp.roll(lambdas_center,  1, axis=0)
        lam2 = lambdas_center
        lam3 = jnp.roll(lambdas_center, -1, axis=0)
        lam4 = jnp.roll(lambdas_center, -2, axis=0)
        lam5 = jnp.roll(lambdas_center, -3, axis=0)
        lam_stack = jnp.stack([lam0, lam1, lam2, lam3, lam4, lam5], axis=0)
        amx = jnp.max(jnp.abs(lam_stack), axis=0)

        # Now use the exact same definitions as original for aterm/bterm/cterm/dterm
        aterm_p = 0.5 * (d0 + amx * dq0)
        bterm_p = 0.5 * (d1 + amx * dq1)
        cterm_p = 0.5 * (d2 + amx * dq2)
        dterm_p = 0.5 * (d3 + amx * dq3)

        IS0_p = 13.0 * (aterm_p - bterm_p)**2 + 3.0 * (aterm_p - 3.0*bterm_p)**2
        IS1_p = 13.0 * (bterm_p - cterm_p)**2 + 3.0 * (bterm_p + cterm_p)**2
        IS2_p = 13.0 * (cterm_p - dterm_p)**2 + 3.0 * (3.0*cterm_p - dterm_p)**2

        alpha0_p = 1.0 / (epsilon + IS0_p)**2
        alpha1_p = 6.0 / (epsilon + IS1_p)**2
        alpha2_p = 3.0 / (epsilon + IS2_p)**2

        alpha_sum_p = alpha0_p + alpha1_p + alpha2_p
        alpha_sum_p = jnp.maximum(alpha_sum_p, 1e-14)  # prevent division by zero

        omega0_p = alpha0_p / alpha_sum_p
        omega2_p = alpha2_p / alpha_sum_p

        second = (omega0_p * (aterm_p - 2.0*bterm_p + cterm_p) / 3.0
                  + (omega2_p - 0.5) * (bterm_p - 2.0*cterm_p + dterm_p) / 6.0)

        # Backward WENO similarly with the matching stencil differences:
        aterm_m = 0.5 * (d4 - amx * dq4)   # corresponds to original dFsk[4] etc.
        bterm_m = 0.5 * (d3 - amx * dq3)
        cterm_m = 0.5 * (d2 - amx * dq2)
        dterm_m = 0.5 * (d1 - amx * dq1)

        IS0_m = 13.0 * (aterm_m - bterm_m)**2 + 3.0 * (aterm_m - 3.0*bterm_m)**2
        IS1_m = 13.0 * (bterm_m - cterm_m)**2 + 3.0 * (bterm_m + cterm_m)**2
        IS2_m = 13.0 * (cterm_m - dterm_m)**2 + 3.0 * (3.0*cterm_m - dterm_m)**2

        alpha0_m = 1.0 / (epsilon + IS0_m)**2
        alpha1_m = 6.0 / (epsilon + IS1_m)**2
        alpha2_m = 3.0 / (epsilon + IS2_m)**2

        alpha_sum_m = alpha0_m + alpha1_m + alpha2_m
        alpha_sum_m = jnp.maximum(alpha_sum_m, 1e-14)  # prevent division by zero

        omega0_m = alpha0_m / alpha_sum_m
        omega2_m = alpha2_m / alpha_sum_m

        third = (omega0_m * (aterm_m - 2.0*bterm_m + cterm_m) / 3.0
                 + (omega2_m - 0.5) * (bterm_m - 2.0*cterm_m + dterm_m) / 6.0)

        Fs = -second + third

        # transform back and add to current flux
        R_col = _eigen_R_col(conserved_state, rhomin, pgmin, gamma, registered_variables, mode)
        dF = jnp.einsum('nxyz,xyz->nxyz', R_col, Fs)
        return F_current + dF
    
    return jax.lax.fori_loop(
        0, 7,
        mode_flux,
        F_interface
    )

@partial(jax.jit, static_argnames=["registered_variables", "config"])
def _weno_flux_y(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
):
    # Transpose to make y the "x" direction
    qy = jnp.transpose(conserved_state, (0, 2, 1, 3))
    
    # Swap components
    momentum_x = qy[registered_variables.momentum_index.x]
    momentum_y = qy[registered_variables.momentum_index.y]
    B_x = qy[registered_variables.magnetic_index.x]
    B_y = qy[registered_variables.magnetic_index.y]
    
    qy = qy.at[registered_variables.momentum_index.x].set(momentum_y)
    qy = qy.at[registered_variables.momentum_index.y].set(momentum_x)
    qy = qy.at[registered_variables.magnetic_index.x].set(B_y)
    qy = qy.at[registered_variables.magnetic_index.y].set(B_x)
    
    Fy = _weno_flux_x(qy, rhomin, pgmin, gamma, config, registered_variables)
    
    # Transpose back
    Fy = jnp.transpose(Fy, (0, 2, 1, 3))
    
    # Swap components back
    Fmomentum_x = Fy[registered_variables.momentum_index.x]
    Fmomentum_y = Fy[registered_variables.momentum_index.y]
    FB_x = Fy[registered_variables.magnetic_index.x]
    FB_y = Fy[registered_variables.magnetic_index.y]
    
    Fy = Fy.at[registered_variables.momentum_index.x].set(Fmomentum_y)
    Fy = Fy.at[registered_variables.momentum_index.y].set(Fmomentum_x)
    Fy = Fy.at[registered_variables.magnetic_index.x].set(FB_y)
    Fy = Fy.at[registered_variables.magnetic_index.y].set(FB_x)
    
    return Fy


@partial(jax.jit, static_argnames=["registered_variables", "config"])
def _weno_flux_z(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
):
    # Transpose to make z the "x" direction
    qz = jnp.transpose(conserved_state, (0, 3, 2, 1))
    
    # Swap components
    momentum_x = qz[registered_variables.momentum_index.x]
    momentum_z = qz[registered_variables.momentum_index.z]
    B_x = qz[registered_variables.magnetic_index.x]
    B_z = qz[registered_variables.magnetic_index.z]
    
    qz = qz.at[registered_variables.momentum_index.x].set(momentum_z)
    qz = qz.at[registered_variables.momentum_index.z].set(momentum_x)
    qz = qz.at[registered_variables.magnetic_index.x].set(B_z)
    qz = qz.at[registered_variables.magnetic_index.z].set(B_x)
    
    Fz = _weno_flux_x(qz, rhomin, pgmin, gamma, config, registered_variables)
    
    # Transpose back
    Fz = jnp.transpose(Fz, (0, 3, 2, 1))
    
    # Swap components back
    Fmomentum_x = Fz[registered_variables.momentum_index.x]
    Fmomentum_z = Fz[registered_variables.momentum_index.z]
    FB_x = Fz[registered_variables.magnetic_index.x]
    FB_z = Fz[registered_variables.magnetic_index.z]
    
    Fz = Fz.at[registered_variables.momentum_index.x].set(Fmomentum_z)
    Fz = Fz.at[registered_variables.momentum_index.z].set(Fmomentum_x)
    Fz = Fz.at[registered_variables.magnetic_index.x].set(FB_z)
    Fz = Fz.at[registered_variables.magnetic_index.z].set(FB_x)
    
    return Fz

# also seems to be slower
# @partial(jax.jit, static_argnames=["registered_variables"])
# def _weno_flux_x(
#     conserved_state,
#     rhomin: Union[float, jnp.ndarray],
#     pgmin: Union[float, jnp.ndarray],
#     gamma: Union[float, jnp.ndarray],
#     registered_variables: RegisteredVariables,
# ):
#     """
#     WENO flux reconstruction.
#     """
    
#     q = conserved_state
#     epsilon = 1e-8
    
#     # Get eigenstructure
#     lambdas_center, R, L = _eigen_x(q, rhomin, pgmin, gamma, registered_variables)
    
#     # Get physical fluxes at cell centers
#     F = _mhd_flux_x(q, gamma, registered_variables)

#     F_stencil = jnp.stack([
#         jnp.roll(F, 2, axis=1),   # i-2
#         jnp.roll(F, 1, axis=1),   # i-1
#         F,                        # i
#         jnp.roll(F, -1, axis=1),  # i+1
#         jnp.roll(F, -2, axis=1),  # i+2
#         jnp.roll(F, -3, axis=1),  # i+3
#     ], axis=0)
    
#     q_stencil = jnp.stack([
#         jnp.roll(q, 2, axis=1),
#         jnp.roll(q, 1, axis=1),
#         q,
#         jnp.roll(q, -1, axis=1),
#         jnp.roll(q, -2, axis=1),
#         jnp.roll(q, -3, axis=1),
#     ], axis=0)
    
#     # Get maximum eigenvalue over stencil
#     lambda_stencil = jnp.stack([
#         jnp.roll(lambdas_center, 2, axis=1),
#         jnp.roll(lambdas_center, 1, axis=1),
#         lambdas_center,
#         jnp.roll(lambdas_center, -1, axis=1),
#         jnp.roll(lambdas_center, -2, axis=1),
#         jnp.roll(lambdas_center, -3, axis=1),
#     ], axis=0)
    
#     amx = jnp.max(jnp.abs(lambda_stencil), axis=0)  # Shape: (7, Nx, Ny, Nz)
    
#     # Transform stencil to characteristic variables
#     # L has shape (7, N_vars, Nx, Ny, Nz) at interfaces
#     # F_stencil: (6, N_vars, Nx, Ny, Nz)
#     # Result: (6, 7, Nx, Ny, Nz)
#     Fsk = jnp.einsum('mnxyz,snxyz->smxyz', L, F_stencil)
#     qsk = jnp.einsum('mnxyz,snxyz->smxyz', L, q_stencil)
    
#     # Compute differences
#     dFsk = Fsk[1:, ...] - Fsk[:-1, ...]  # (5, 7, Nx, Ny, Nz)
#     dqsk = qsk[1:, ...] - qsk[:-1, ...]
    
#     # Forward WENO (positive flux)
#     aterm_p = 0.5 * (dFsk[0, ...] + amx * dqsk[0, ...])
#     bterm_p = 0.5 * (dFsk[1, ...] + amx * dqsk[1, ...])
#     cterm_p = 0.5 * (dFsk[2, ...] + amx * dqsk[2, ...])
#     dterm_p = 0.5 * (dFsk[3, ...] + amx * dqsk[3, ...])
    
#     IS0_p = 13.0 * (aterm_p - bterm_p)**2 + 3.0 * (aterm_p - 3.0*bterm_p)**2
#     IS1_p = 13.0 * (bterm_p - cterm_p)**2 + 3.0 * (bterm_p + cterm_p)**2
#     IS2_p = 13.0 * (cterm_p - dterm_p)**2 + 3.0 * (3.0*cterm_p - dterm_p)**2
    
#     alpha0_p = 1.0 / (epsilon + IS0_p)**2
#     alpha1_p = 6.0 / (epsilon + IS1_p)**2
#     alpha2_p = 3.0 / (epsilon + IS2_p)**2
    
#     alpha_sum_p = alpha0_p + alpha1_p + alpha2_p
#     alpha_sum_p = jnp.maximum(alpha_sum_p, 1e-14)  # prevent division by zero

#     omega0_p = alpha0_p / alpha_sum_p
#     omega2_p = alpha2_p / alpha_sum_p
    
#     second = (omega0_p * (aterm_p - 2.0*bterm_p + cterm_p) / 3.0 
#               + (omega2_p - 0.5) * (bterm_p - 2.0*cterm_p + dterm_p) / 6.0)
    
#     # Backward WENO (negative flux)
#     aterm_m = 0.5 * (dFsk[4, ...] - amx * dqsk[4, ...])
#     bterm_m = 0.5 * (dFsk[3, ...] - amx * dqsk[3, ...])
#     cterm_m = 0.5 * (dFsk[2, ...] - amx * dqsk[2, ...])
#     dterm_m = 0.5 * (dFsk[1, ...] - amx * dqsk[1, ...])
    
#     IS0_m = 13.0 * (aterm_m - bterm_m)**2 + 3.0 * (aterm_m - 3.0*bterm_m)**2
#     IS1_m = 13.0 * (bterm_m - cterm_m)**2 + 3.0 * (bterm_m + cterm_m)**2
#     IS2_m = 13.0 * (cterm_m - dterm_m)**2 + 3.0 * (3.0*cterm_m - dterm_m)**2
    
#     alpha0_m = 1.0 / (epsilon + IS0_m)**2
#     alpha1_m = 6.0 / (epsilon + IS1_m)**2
#     alpha2_m = 3.0 / (epsilon + IS2_m)**2
    
#     alpha_sum_m = alpha0_m + alpha1_m + alpha2_m
#     alpha_sum_m = jnp.maximum(alpha_sum_m, 1e-14)  # prevent division by zero

#     omega0_m = alpha0_m / alpha_sum_m
#     omega2_m = alpha2_m / alpha_sum_m
    
#     third = (omega0_m * (aterm_m - 2.0*bterm_m + cterm_m) / 3.0 
#              + (omega2_m - 0.5) * (bterm_m - 2.0*cterm_m + dterm_m) / 6.0)
    
#     # Combine
#     Fs = - second + third  # Shape: (7, Nx, Ny, Nz)
    
#     # Transform back to physical variables
#     # R has shape (N_vars, 7, Nx, Ny, Nz)
#     dF = jnp.einsum('nmxyz,mxyz->nxyz', R, Fs)  # Shape: (N_vars, Nx, Ny, Nz)

#     # Base 4th-order flux
#     # first = (-Fsk[1, ...] + 7.0*Fsk[2, ...] + 7.0*Fsk[3, ...] - Fsk[4, ...]) / 12.0
#     first = 1/12 * (
#         -jnp.roll(F, 1, axis=1) + 7 * F + 7 * jnp.roll(F, -1, axis=1) - jnp.roll(F, -2, axis=1)
#     )

#     dF = first + dF

#     return dF


import jax
import jax.numpy as jnp


@jax.jit
def cross(a, b):
    result = jnp.zeros_like(a)
    result = result.at[0, ...].set(a[1, ...] * b[2, ...] - a[2, ...] * b[1, ...])
    result = result.at[1, ...].set(a[2, ...] * b[0, ...] - a[0, ...] * b[2, ...])
    result = result.at[2, ...].set(a[0, ...] * b[1, ...] - a[1, ...] * b[0, ...])
    return result


@jax.jit
def divergence2D(field, grid_spacing: float):
    """Calculate the divergence of a 3-component field on a 2D grid.
    field.shape == (3, nx, ny)
    """
    divergence = jnp.zeros_like(field)

    # d/dx of component 0: roll along x -> axis 0 of field[0]
    dfx = (jnp.roll(field[0], -1, axis=0) - jnp.roll(field[0], 1, axis=0)) / (2 * grid_spacing)
    divergence = divergence.at[0].add(dfx)

    # d/dy of component 1: roll along y -> axis 1 of field[1]
    dfy = (jnp.roll(field[1], -1, axis=1) - jnp.roll(field[1], 1, axis=1)) / (2 * grid_spacing)
    divergence = divergence.at[1].add(dfy)

    return jnp.sum(divergence, axis=0)


@jax.jit
def divergence3D(field, grid_spacing: float):
    """Calculate the divergence of a 3-component field on a 3D grid.
    field.shape == (3, nx, ny, nz)
    """
    divergence = jnp.zeros_like(field)

    # d/dx of component 0: roll along x -> axis 0 of field[0]
    dfx = (jnp.roll(field[0], -1, axis=0) - jnp.roll(field[0], 1, axis=0)) / (2 * grid_spacing)
    divergence = divergence.at[0].add(dfx)

    # d/dy of component 1: roll along y -> axis 1 of field[1]
    dfy = (jnp.roll(field[1], -1, axis=1) - jnp.roll(field[1], 1, axis=1)) / (2 * grid_spacing)
    divergence = divergence.at[1].add(dfy)

    # d/dz of component 2: roll along z -> axis 2 of field[2]
    dfz = (jnp.roll(field[2], -1, axis=2) - jnp.roll(field[2], 1, axis=2)) / (2 * grid_spacing)
    divergence = divergence.at[2].add(dfz)

    return jnp.sum(divergence, axis=0)


@jax.jit
def curl3D(field, grid_spacing: float):
    """Calculate the curl of a 3-component field on a 3D grid.
    field.shape == (3, nx, ny, nz)
    """
    curl = jnp.zeros_like(field)

    # curl_x = dFz/dy - dFy/dz
    dFz_dy = 0.5 * (jnp.roll(field[2], -1, axis=1) - jnp.roll(field[2], 1, axis=1)) / grid_spacing
    dFy_dz = 0.5 * (jnp.roll(field[1], -1, axis=2) - jnp.roll(field[1], 1, axis=2)) / grid_spacing
    curl = curl.at[0].add(dFz_dy - dFy_dz)

    # curl_y = dFx/dz - dFz/dx
    dFx_dz = 0.5 * (jnp.roll(field[0], -1, axis=2) - jnp.roll(field[0], 1, axis=2)) / grid_spacing
    dFz_dx = 0.5 * (jnp.roll(field[2], -1, axis=0) - jnp.roll(field[2], 1, axis=0)) / grid_spacing
    curl = curl.at[1].add(dFx_dz - dFz_dx)

    # curl_z = dFy/dx - dFx/dy
    dFy_dx = 0.5 * (jnp.roll(field[1], -1, axis=0) - jnp.roll(field[1], 1, axis=0)) / grid_spacing
    dFx_dy = 0.5 * (jnp.roll(field[0], -1, axis=1) - jnp.roll(field[0], 1, axis=1)) / grid_spacing
    curl = curl.at[2].add(dFy_dx - dFx_dy)

    return curl


@jax.jit
def curl2D(field, grid_spacing: float):
    """Calculate the curl of a 3-component field on a 2D grid.
    field.shape == (3, nx, ny)
    (assumes grid_spacing is dy in some of your original docstring contexts)
    """
    curl = jnp.zeros_like(field)

    # curl_x = dFz/dy  (roll y -> axis=1 of field[2])
    dFz_dy = 0.5 * (jnp.roll(field[2], -1, axis=1) - jnp.roll(field[2], 1, axis=1)) / grid_spacing
    curl = curl.at[0].add(dFz_dy)

    # curl_y = - dFz/dx  (roll x -> axis=0 of field[2])
    dFz_dx = 0.5 * (jnp.roll(field[2], -1, axis=0) - jnp.roll(field[2], 1, axis=0)) / grid_spacing
    curl = curl.at[1].add(-dFz_dx)

    # curl_z = dFy/dx - dFx/dy
    dFy_dx = 0.5 * (jnp.roll(field[1], -1, axis=0) - jnp.roll(field[1], 1, axis=0)) / grid_spacing
    dFx_dy = 0.5 * (jnp.roll(field[0], -1, axis=1) - jnp.roll(field[0], 1, axis=1)) / grid_spacing
    curl = curl.at[2].add(dFy_dx - dFx_dy)

    return curl



# general
import jax.numpy as jnp
import jax
from functools import partial

# type checking
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union
from astronomix._physics_modules._stellar_wind.stellar_wind import _wind_injection
from astronomix.option_classes.simulation_config import STATE_TYPE, UNSPLIT

# astronomix containers
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.option_classes.simulation_config import SimulationConfig
from astronomix.option_classes.simulation_params import SimulationParams
from astronomix.variable_registry.registered_variables import RegisteredVariables

# astronomix functions
from astronomix._fluid_equations._fluxes import _euler_flux
from astronomix._fluid_equations._equations import speed_of_sound
from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import (
    gas_pressure_from_primitives_with_crs,
    speed_of_sound_crs,
)
from astronomix._physics_modules.run_physics_modules import _run_physics_modules

# better use same as in the riemann solver??
# now these wave speeds are calculated without
# the reconstruction - for the purely spatial
# reconstruction, we do not need to know
# the time step a priori


# TODO: merge duplicate code in this and hll.py
# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["registered_variables", "config", "flux_direction_index"]
)
def get_wave_speeds(
    primitives_left: STATE_TYPE,
    primitives_right: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
    config: SimulationConfig,
    flux_direction_index: int,
) -> Union[float, Float[Array, ""]]:
    """
    Returns the conservative fluxes.

    Args:
        primitives_left: States left of the interfaces.
        primitives_right: States right of the interfaces.
        gamma: The adiabatic index.

    Returns:
        The conservative fluxes at the interfaces.

    """

    rho_L = primitives_left[registered_variables.density_index]
    u_L = primitives_left[flux_direction_index]

    rho_R = primitives_right[registered_variables.density_index]
    u_R = primitives_right[flux_direction_index]

    p_L = primitives_left[registered_variables.pressure_index]
    p_R = primitives_right[registered_variables.pressure_index]

    # calculate the sound speeds
    if not config.cosmic_ray_config.cosmic_rays:
        c_L = speed_of_sound(rho_L, p_L, gamma)
        c_R = speed_of_sound(rho_R, p_R, gamma)
    else:
        c_L = speed_of_sound_crs(primitives_left, registered_variables)
        c_R = speed_of_sound_crs(primitives_right, registered_variables)

    # very simple approach for the wave velocities
    # wave_speeds_right_plus = jnp.maximum(jnp.maximum(u_L + c_L, u_R + c_R), 0)
    # wave_speeds_left_minus = jnp.minimum(jnp.minimum(u_L - c_L, u_R - c_R), 0)

    wave_speeds_right_plus = jnp.abs(u_L) + c_L
    wave_speeds_left_minus = jnp.abs(u_R) + c_R

    max_wave_speed = jnp.maximum(
        jnp.max(jnp.abs(wave_speeds_right_plus)),
        jnp.max(jnp.abs(wave_speeds_left_minus)),
    )

    return max_wave_speed


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _cfl_time_step(
    primitive_state: STATE_TYPE,
    grid_spacing: Union[float, Float[Array, ""]],
    dt_max: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    C_CFL: Union[float, Float[Array, ""]] = 0.8,
) -> Float[Array, ""]:
    """Calculate the time step based on the CFL condition.

    Args:
        primitive_state: The primitive state array.
        grid_spacing: The cell width.
        dt_max: The maximum time step.
        gamma: The adiabatic index.
        C_CFL: The CFL number.

    Returns:
        The time step.

    """

    if config.split == UNSPLIT:
        rho = primitive_state[registered_variables.density_index]
        p = primitive_state[registered_variables.pressure_index]
        c = speed_of_sound(rho, p, gamma)
        alpha_lax = jnp.zeros((config.dimensionality,))
        for axis in range(1, config.dimensionality + 1):
            u = primitive_state[axis]
            alpha_lax_i = jnp.max(jnp.abs(u) + c)
            alpha_lax = alpha_lax.at[axis - 1].set(alpha_lax_i)

        return C_CFL * 1 / jnp.sum(alpha_lax / grid_spacing)

    else:
        if config.dimensionality == 3:
            # wave speeds in x direction
            primitive_state_left = primitive_state[:, :-1, :, :]
            primitive_state_right = primitive_state[:, 1:, :, :]
            max_wave_speed_x = get_wave_speeds(
                primitive_state_left,
                primitive_state_right,
                gamma,
                registered_variables,
                config,
                registered_variables.velocity_index.x,
            )

            # wave speeds in y direction
            primitive_state_left = primitive_state[:, :, :-1, :]
            primitive_state_right = primitive_state[:, :, 1:, :]
            max_wave_speed_y = get_wave_speeds(
                primitive_state_left,
                primitive_state_right,
                gamma,
                registered_variables,
                config,
                registered_variables.velocity_index.y,
            )

            # wave speeds in z direction
            primitive_state_left = primitive_state[:, :, :, :-1]
            primitive_state_right = primitive_state[:, :, :, 1:]
            max_wave_speed_z = get_wave_speeds(
                primitive_state_left,
                primitive_state_right,
                gamma,
                registered_variables,
                config,
                registered_variables.velocity_index.z,
            )

            # get the maximum wave speed
            max_wave_speed = jnp.maximum(
                jnp.maximum(max_wave_speed_x, max_wave_speed_y), max_wave_speed_z
            )
        elif config.dimensionality == 2:
            # wave speeds in x direction
            primitive_state_left = primitive_state[:, :-1, :]
            primitive_state_right = primitive_state[:, 1:, :]
            max_wave_speed_x = get_wave_speeds(
                primitive_state_left,
                primitive_state_right,
                gamma,
                registered_variables,
                config,
                registered_variables.velocity_index.x,
            )

            # wave speeds in y direction
            primitive_state_left = primitive_state[:, :, :-1]
            primitive_state_right = primitive_state[:, :, 1:]
            max_wave_speed_y = get_wave_speeds(
                primitive_state_left,
                primitive_state_right,
                gamma,
                registered_variables,
                config,
                registered_variables.velocity_index.y,
            )

            # get the maximum wave speed
            max_wave_speed = jnp.maximum(max_wave_speed_x, max_wave_speed_y)
        else:
            # wave speeds in x direction
            primitive_state_left = primitive_state[:, :-1]
            primitive_state_right = primitive_state[:, 1:]
            max_wave_speed = get_wave_speeds(
                primitive_state_left,
                primitive_state_right,
                gamma,
                registered_variables,
                config,
                registered_variables.velocity_index,
            )

        # calculate the time step
        dt = C_CFL * grid_spacing / max_wave_speed

        if config.use_max_adaptive_timestep:
            dt = jnp.minimum(dt, dt_max)
        
        return dt


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _source_term_aware_time_step(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
    current_time: Union[float, Float[Array, ""]],
) -> Float[Array, ""]:
    """
    Calculate the time step based on the CFL condition and the source terms. What timestep
    would be chosen if the source terms were added under the current CFL time step?

    Args:
        state: The state array.
        config: The configuration.
        params: The parameters.
        helper_data: The helper data.

    Returns:
        The time step.
    """

    # == experimental: correct the CFL time step based on the physical sources ==

    # calculate the time step based on the CFL condition
    dt = _cfl_time_step(
        primitive_state,
        config.grid_spacing,
        params.dt_max,
        params.gamma,
        config,
        registered_variables,
        params.C_cfl,
    )

    # independent of config.use_max_adaptive_timestep
    dt = jnp.minimum(dt, params.dt_max)

    # ONLY ADD THE STELLAR WIND SOURCE TERM
    # DO NOT RUN ALL PHYSICS MODULES HERE

    # TODO: HOW TO HANDLE FURTHER SOURCE TERMS?

    hypothetical_new_state = _wind_injection(
        primitive_state, dt, config, params, helper_data, registered_variables
    )

    # hypothetical_new_state = _run_physics_modules(
    #     primitive_state,
    #     dt,
    #     config,
    #     params,
    #     helper_data,
    #     registered_variables,
    #     current_time,
    # )

    dt = _cfl_time_step(
        hypothetical_new_state,
        config.grid_spacing,
        params.dt_max,
        params.gamma,
        config,
        registered_variables,
        params.C_cfl,
    )

    # ===========================================================================

    return dt



"""
Here we protect the density and pressure from going negative.

In my view this is a bit of a shady practice, hiding unphysical
updates under the rug. However, it is common practice.
"""

import jax.numpy as jnp
import jax
from functools import partial

from jaxtyping import Array, Float

from typing import Union

from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    STATE_TYPE,
)

@partial(
    jax.jit, static_argnames=["registered_variables"]
)
def _enforce_positivity(
    conserved_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    minimum_density: Union[float, Float[Array, ""]],
    minimum_pressure: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    

    rho = conserved_state[registered_variables.density_index]

    # enforce minimum density
    rho = jnp.maximum(rho, minimum_density)

    v_x = conserved_state[registered_variables.momentum_index.x] / rho
    v_y = conserved_state[registered_variables.momentum_index.y] / rho
    v_z = conserved_state[registered_variables.momentum_index.z] / rho
    energy = conserved_state[registered_variables.energy_index]
    B_x = conserved_state[registered_variables.magnetic_index.x]
    B_y = conserved_state[registered_variables.magnetic_index.y]
    B_z = conserved_state[registered_variables.magnetic_index.z]

    b2 = B_x**2 + B_y**2 + B_z**2
    v2 = v_x**2 + v_y**2 + v_z**2

    # calculate pressure
    pressure = (gamma - 1.0) * (energy - 0.5 * rho * v2 - 0.5 * b2)
    pressure = jnp.maximum(pressure, minimum_pressure)

    # redefine energy with new pressure
    energy = pressure / (gamma - 1.0) + 0.5 * rho * v2 + 0.5 * b2

    # reconstruct conserved state
    conserved_state = conserved_state.at[registered_variables.density_index].set(rho)
    conserved_state = conserved_state.at[registered_variables.energy_index].set(energy)

    return conserved_state


from functools import partial
from types import NoneType
from typing import NamedTuple, Union
from jax import NamedSharding
import jax.numpy as jnp
from astronomix._geometry.geometry import _center_of_volume, _r_hat_alpha
from astronomix.option_classes.simulation_config import (
    CARTESIAN,
    CYLINDRICAL,
    SPHERICAL,
    SimulationConfig,
)
import jax

# Helper data like the radii and cell volumes
# in the simulation or cooling tables etc.


class HelperData(NamedTuple):
    """Helper data used throughout the simulation."""

    #: The geometric centers of the cells.
    geometric_centers: jnp.ndarray = None

    #: The volumetric centers of the cells.
    #: Same as the geometric centers for Cartesian geometry.
    volumetric_centers: jnp.ndarray = None

    #: cell center to box center distances
    #: only for config.dimensionality > 1
    r: jnp.ndarray = None

    #: A helper variable, defined as
    #: \hat{r}^\alpha = V_j / (2 * \alpha * \pi * \Delta r)
    #: with V_j the volume of cell j, \alpha the geometry factor
    #: and \Delta r the cell width.
    r_hat_alpha: jnp.ndarray = None

    #: The cell volumes.
    cell_volumes: jnp.ndarray = None

    #: Coordinates of the inner cell boundaries.
    inner_cell_boundaries: jnp.ndarray = None

    #: Coordinates of the outer cell boundaries.
    outer_cell_boundaries: jnp.ndarray = None

@partial(jax.jit, static_argnames=("config", "sharding", "padded", "production"))
def get_helper_data(
    config: SimulationConfig,
    sharding: Union[NoneType, NamedSharding] = None,
    padded: bool = False,
    production: bool = False,
) -> HelperData:
    """Generate the helper data for the simulation from the configuration."""

    if padded:
        ngc = config.num_ghost_cells
    else:
        ngc = 0

    grid_spacing = config.box_size / config.num_cells

    # in spherical or cylindrical symmetry, we always need the helper data
    if config.geometry == SPHERICAL or config.geometry == CYLINDRICAL:
        r = jnp.linspace(
            grid_spacing / 2 - ngc * grid_spacing,
            config.box_size + grid_spacing / 2 + ngc * grid_spacing,
            config.num_cells + 2 * ngc,
            endpoint=False,
        )
        inner_cell_boundaries = r - grid_spacing / 2
        outer_cell_boundaries = r + grid_spacing / 2
        volumetric_centers = _center_of_volume(r, grid_spacing, config.geometry)
        r_hat = _r_hat_alpha(r, grid_spacing, config.geometry)
        cell_volumes = 2 * config.geometry * jnp.pi * grid_spacing * r_hat
        helper_data_pad = HelperData(
            geometric_centers=r,
            volumetric_centers=volumetric_centers,
            r_hat_alpha=r_hat,
            cell_volumes=cell_volumes,
            inner_cell_boundaries=inner_cell_boundaries,
            outer_cell_boundaries=outer_cell_boundaries,
        )

    helper_data_necessary = (
        config.wind_config.stellar_wind or config.cooling_config.cooling or 
        config.return_snapshots
    )

    if not production or helper_data_necessary:
        
        if config.geometry == CARTESIAN:
            if config.dimensionality > 1:
                x = jnp.linspace(
                    grid_spacing / 2 - ngc * grid_spacing,
                    config.box_size + grid_spacing / 2 + ngc * grid_spacing,
                    config.num_cells + 2 * ngc,
                    endpoint=False,
                )
                y = jnp.linspace(
                    grid_spacing / 2 - ngc * grid_spacing,
                    config.box_size + grid_spacing / 2 + ngc * grid_spacing,
                    config.num_cells + 2 * ngc,
                    endpoint=False,
                )

                if config.dimensionality == 3:
                    z = jnp.linspace(
                        grid_spacing / 2 - ngc * grid_spacing,
                        config.box_size + grid_spacing / 2 + ngc * grid_spacing,
                        config.num_cells + 2 * ngc,
                        endpoint=False,
                    )
                    if sharding is not None:
                        geometric_centers = jax.lax.with_sharding_constraint(
                            jnp.array(jnp.meshgrid(x, y, z)), sharding
                        )
                    else:
                        geometric_centers = jnp.array(jnp.meshgrid(x, y, z))
                else:
                    geometric_centers = jnp.array(jnp.meshgrid(x, y))

                # calculate the distances from the cell centers to the box center
                box_center = jnp.zeros(config.dimensionality) + config.box_size / 2

                geometric_centers = jnp.moveaxis(geometric_centers, 0, -1)

                volumetric_centers = geometric_centers

                r = jnp.linalg.norm(geometric_centers - box_center, axis=-1)

                helper_data_pad = HelperData(
                    geometric_centers=geometric_centers,
                    volumetric_centers=volumetric_centers,
                    r=r,
                )
            else:
                r = jnp.linspace(
                    grid_spacing / 2 - ngc * grid_spacing,
                    config.box_size - grid_spacing / 2 + ngc * grid_spacing,
                    config.num_cells + 2 * ngc,
                )
                r_hat = grid_spacing * jnp.ones_like(r)  # not really
                cell_volumes = grid_spacing * jnp.ones_like(r)
                inner_cell_boundaries = r - grid_spacing / 2
                outer_cell_boundaries = r + grid_spacing / 2
                helper_data_pad = HelperData(
                    geometric_centers=r,
                    r_hat_alpha=r_hat,
                    cell_volumes=cell_volumes,
                    inner_cell_boundaries=inner_cell_boundaries,
                    outer_cell_boundaries=outer_cell_boundaries,
                    volumetric_centers=r,
                )
    
    else:
        helper_data_pad = HelperData()

    return helper_data_pad



"""
Computations of the eigenvalues and eigenvectors for the MHD equations.

The eigenstructure was extracted from the HOW-MHD Fortran code, with altered
variable names for clarity and altered numerical safeguards.

NOTE: Problems for differentiation largely follow from square roots and divisions:
The derivative of sqrt(x) is 1/(2*sqrt(x)) and of 1/x is -1/x^2, where both expressions
are problematic for small x, especially when multiplying gradients in the backward pass,
-> exploding gradients.
"""

from functools import partial
import jax
import jax.numpy as jnp
from typing import Union

from astronomix.variable_registry.registered_variables import RegisteredVariables


def diff_safe_sqrt(x):
    epsilon = 1e-12
    x_safe = jnp.maximum(x, epsilon)
    return jnp.sqrt(x_safe)


@partial(jax.jit, static_argnames=["registered_variables"])
def _eigenvalue_building_blocks(
    conserved_state,
    gamma,
    rhomin,
    pgmin,
    registered_variables: RegisteredVariables,
):
    # unpack the conserved variables
    density = conserved_state[registered_variables.density_index]
    momentum_x = conserved_state[registered_variables.momentum_index.x]
    momentum_y = conserved_state[registered_variables.momentum_index.y]
    momentum_z = conserved_state[registered_variables.momentum_index.z]
    magnetic_x = conserved_state[registered_variables.magnetic_index.x]
    magnetic_y = conserved_state[registered_variables.magnetic_index.y]
    magnetic_z = conserved_state[registered_variables.magnetic_index.z]
    energy = conserved_state[registered_variables.energy_index]

    # compute primitives
    rho = density
    velocity_x = momentum_x / rho
    velocity_y = momentum_y / rho
    velocity_z = momentum_z / rho
    velocity_squared = (
        velocity_x * velocity_x + velocity_y * velocity_y + velocity_z * velocity_z
    )
    magnetic_field_squared = (
        magnetic_x * magnetic_x + magnetic_y * magnetic_y + magnetic_z * magnetic_z
    )
    gas_pressure = (gamma - 1.0) * (
        energy - 0.5 * (rho * velocity_squared + magnetic_field_squared)
    )

    # redefine the density and pressure, and energy based on floors
    rho = jnp.where(
        (rho < rhomin) | (gas_pressure < pgmin), jnp.maximum(rho, rhomin), rho
    )
    gas_pressure = jnp.where(
        (rho < rhomin) | (gas_pressure < pgmin),
        jnp.maximum(gas_pressure, pgmin),
        gas_pressure,
    )
    energy = jnp.where(
        (rho < rhomin) | (gas_pressure < pgmin),
        gas_pressure / (gamma - 1.0)
        + 0.5 * (rho * velocity_squared + magnetic_field_squared),
        energy,
    )

    # compute derived quantities
    sound_speed_sq = jnp.maximum(0.0, gamma * jnp.abs(gas_pressure / rho))
    magnetosonic_discriminant_root = diff_safe_sqrt(
        jnp.maximum(
            0.0,
            (magnetic_field_squared / rho + sound_speed_sq) ** 2
            - 4.0 * (magnetic_x * magnetic_x) / rho * sound_speed_sq,
        )
    )
    fast_magnetosonic_velocity = diff_safe_sqrt(
        jnp.maximum(
            0.0,
            0.5
            * (
                magnetic_field_squared / rho
                + sound_speed_sq
                + magnetosonic_discriminant_root
            ),
        )
    )
    alfven_velocity = diff_safe_sqrt(jnp.maximum(0.0, (magnetic_x * magnetic_x) / rho))
    slow_magnetosonic_velocity = diff_safe_sqrt(
        jnp.maximum(
            0.0,
            0.5
            * (
                magnetic_field_squared / rho
                + sound_speed_sq
                - magnetosonic_discriminant_root
            ),
        )
    )

    return (
        velocity_x,
        fast_magnetosonic_velocity,
        alfven_velocity,
        slow_magnetosonic_velocity,
    )

@partial(jax.jit, static_argnames=["registered_variables"])
def _eigenvector_building_blocks(
    conserved_state,
    gamma,
    rhomin,
    pgmin,
    registered_variables: RegisteredVariables,
):
    # unpack conserved variables
    rho = conserved_state[registered_variables.density_index]  # density
    momentum_x = conserved_state[registered_variables.momentum_index.x]
    momentum_y = conserved_state[registered_variables.momentum_index.y]
    momentum_z = conserved_state[registered_variables.momentum_index.z]
    magnetic_x = conserved_state[registered_variables.magnetic_index.x]
    magnetic_y = conserved_state[registered_variables.magnetic_index.y]
    magnetic_z = conserved_state[registered_variables.magnetic_index.z]
    energy = conserved_state[registered_variables.energy_index]

    # compute primitives
    rho = rho
    velocity_x = momentum_x / rho
    velocity_y = momentum_y / rho
    velocity_z = momentum_z / rho
    velocity_sq = (
        velocity_x * velocity_x + velocity_y * velocity_y + velocity_z * velocity_z
    )
    magnetic_sq = (
        magnetic_x * magnetic_x + magnetic_y * magnetic_y + magnetic_z * magnetic_z
    )
    gas_pressure = (gamma - 1.0) * (energy - 0.5 * (rho * velocity_sq + magnetic_sq))

    # redefine the density and pressure, and energy based on floors
    rho = jnp.where(
        (rho < rhomin) | (gas_pressure < pgmin), jnp.maximum(rho, rhomin), rho
    )
    gas_pressure = jnp.where(
        (rho < rhomin) | (gas_pressure < pgmin),
        jnp.maximum(gas_pressure, pgmin),
        gas_pressure,
    )
    energy = jnp.where(
        (rho < rhomin) | (gas_pressure < pgmin),
        gas_pressure / (gamma - 1.0) + 0.5 * (rho * velocity_sq + magnetic_sq),
        energy,
    )

    specific_enthalpy = (energy + gas_pressure) / rho

    # periodic average to interfaces
    def avg_x(arr):
        return 0.5 * (arr + jnp.roll(arr, shift=-1, axis=0))

    rho_interface = avg_x(rho)
    velocity_x_interface = avg_x(velocity_x)
    velocity_y_interface = avg_x(velocity_y)
    velocity_z_interface = avg_x(velocity_z)
    magnetic_x_interface = avg_x(magnetic_x)
    magnetic_y_interface = avg_x(magnetic_y)
    magnetic_z_interface = avg_x(magnetic_z)
    specific_enthalpy_interface = avg_x(specific_enthalpy)

    # interface derived quantities
    velocity_sq_interface = (
        velocity_x_interface * velocity_x_interface
        + velocity_y_interface * velocity_y_interface
        + velocity_z_interface * velocity_z_interface
    )
    magnetic_sq_interface = (
        magnetic_x_interface * magnetic_x_interface
        + magnetic_y_interface * magnetic_y_interface
        + magnetic_z_interface * magnetic_z_interface
    )
    b_sq_over_rho_interface = magnetic_sq_interface / rho_interface
    bx_sq_over_rho_interface = (
        magnetic_x_interface * magnetic_x_interface
    ) / rho_interface

    # enthalpy based sound speed at interfaces
    sound_speed_sq_interface = (gamma - 1.0) * (
        specific_enthalpy_interface
        - 0.5 * (velocity_sq_interface + b_sq_over_rho_interface)
    )
    sound_speed_interface = diff_safe_sqrt(jnp.maximum(0.0, sound_speed_sq_interface))

    # calculate the characteristic velocities at the interfaces
    magnetosonic_discriminant_interface = (
        b_sq_over_rho_interface + sound_speed_sq_interface
    ) ** 2 - 4.0 * bx_sq_over_rho_interface * sound_speed_sq_interface
    magnetosonic_discriminant_root_interface = diff_safe_sqrt(
        jnp.maximum(0.0, magnetosonic_discriminant_interface)
    )

    fast_magnetosonic_velocity_interface = diff_safe_sqrt(
        jnp.maximum(
            0.0,
            0.5
            * (
                b_sq_over_rho_interface
                + sound_speed_sq_interface
                + magnetosonic_discriminant_root_interface
            ),
        )
    )
    alfven_velocity_interface = diff_safe_sqrt(
        jnp.maximum(0.0, bx_sq_over_rho_interface)
    )
    slow_magnetosonic_velocity_interface = diff_safe_sqrt(
        jnp.maximum(
            0.0,
            0.5
            * (
                b_sq_over_rho_interface
                + sound_speed_sq_interface
                - magnetosonic_discriminant_root_interface
            ),
        )
    )

    # retrieve tangential magnetic field components
    b_tangential_sq = (
        magnetic_y_interface * magnetic_y_interface
        + magnetic_z_interface * magnetic_z_interface
    )
    sgn_bx = jnp.where(magnetic_x_interface >= 0.0, 1.0, -1.0)

    b_tangential_sq = jnp.maximum(b_tangential_sq, 1.0e-12)

    # B_y / (sqrt(B_y^2 + B_z^2))
    bt_normalized_y = jnp.where(
        b_tangential_sq >= 1.0e-12,
        magnetic_y_interface / diff_safe_sqrt(b_tangential_sq),
        1.0 / diff_safe_sqrt(2.0),
    )
    # B_z / (sqrt(B_y^2 + B_z^2))
    bt_normalized_z = jnp.where(
        b_tangential_sq >= 1.0e-12,
        magnetic_z_interface / diff_safe_sqrt(b_tangential_sq),
        1.0 / diff_safe_sqrt(2.0),
    )

    # fast_mode_weighting = sqrt( c_s^2 − λ_slow^2 ) / sqrt( λ_fast^2 − λ_slow^2 )
    # slow_mode_weighting = sqrt( λ_fast^2 − c_s^2 ) / sqrt( λ_fast^2 − λ_slow^2 )
    # these are designed such that fast_mode_weighting^2 + slow_mode_weighting^2 = 1
    denom = (
        fast_magnetosonic_velocity_interface * fast_magnetosonic_velocity_interface
        - slow_magnetosonic_velocity_interface * slow_magnetosonic_velocity_interface
    )
    denom = jnp.maximum(denom, 1.0e-12)
    fast_mode_weighting = jnp.where(
        denom >= 1.0e-12,
        diff_safe_sqrt(
            jnp.maximum(
                0.0,
                sound_speed_sq_interface
                - slow_magnetosonic_velocity_interface
                * slow_magnetosonic_velocity_interface,
            )
        )
        / diff_safe_sqrt(denom),
        1.0,
    )
    slow_mode_weighting = jnp.where(
        denom >= 1e-12,
        diff_safe_sqrt(
            jnp.maximum(
                0.0,
                fast_magnetosonic_velocity_interface
                * fast_magnetosonic_velocity_interface
                - sound_speed_sq_interface,
            )
        )
        / diff_safe_sqrt(denom),
        1.0,
    )

    sqrt_rho = diff_safe_sqrt(rho_interface)

    gam0 = 1.0 - gamma
    gam1 = 0.5 * (gamma - 1.0)
    gam2 = (gamma - 2.0) / (gamma - 1.0)

    sound_speed_sq_inverse = jnp.where(
        sound_speed_sq_interface > 0.0, 1.0 / sound_speed_sq_interface, 0.0
    )

    sgn_bt = jnp.where(
        magnetic_y_interface != 0.0,
        jnp.where(magnetic_y_interface >= 0.0, 1.0, -1.0),
        jnp.where(magnetic_z_interface >= 0.0, 1.0, -1.0),
    )

    sound_speed_greater_alfven_speed = (
        sound_speed_interface >= alfven_velocity_interface
    )

    return (
        rho_interface,
        sqrt_rho,
        velocity_x_interface,
        velocity_y_interface,
        velocity_z_interface,
        velocity_sq_interface,
        magnetic_x_interface,
        magnetic_y_interface,
        magnetic_z_interface,
        bt_normalized_y,
        bt_normalized_z,
        sgn_bx,
        sgn_bt,
        sound_speed_interface,
        sound_speed_sq_interface,
        sound_speed_sq_inverse,
        sound_speed_greater_alfven_speed,
        fast_magnetosonic_velocity_interface,
        alfven_velocity_interface,
        slow_magnetosonic_velocity_interface,
        fast_mode_weighting,
        slow_mode_weighting,
        gam0,
        gam1,
        gam2,
    )


@partial(jax.jit, static_argnames=["registered_variables"])
def _eigen_R_col(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    registered_variables: RegisteredVariables,
    col: int,
):
    (
        rho_interface,
        sqrt_rho,
        velocity_x_interface,
        velocity_y_interface,
        velocity_z_interface,
        velocity_sq_interface,
        magnetic_x_interface,
        magnetic_y_interface,
        magnetic_z_interface,
        bt_normalized_y,
        bt_normalized_z,
        sgn_bx,
        sgn_bt,
        sound_speed_interface,
        sound_speed_sq_interface,
        sound_speed_sq_inverse,
        sound_speed_greater_alfven_speed,
        fast_magnetosonic_velocity_interface,
        alfven_velocity_interface,
        slow_magnetosonic_velocity_interface,
        fast_mode_weighting,
        slow_mode_weighting,
        gam0,
        gam1,
        gam2,
    ) = _eigenvector_building_blocks(
        conserved_state,
        gamma,
        rhomin,
        pgmin,
        registered_variables,
    )

    # shorter names for registry indices
    density_index = registered_variables.density_index
    momentum_index_x = registered_variables.momentum_index.x
    momentum_index_y = registered_variables.momentum_index.y
    momentum_index_z = registered_variables.momentum_index.z
    magnetic_index_y = registered_variables.magnetic_index.y
    magnetic_index_z = registered_variables.magnetic_index.z
    energy_index = registered_variables.energy_index

    def col_0():
        # Column 1 (fast -)
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(fast_mode_weighting)
        R = R.at[momentum_index_x].set(
            fast_mode_weighting
            * (velocity_x_interface - fast_magnetosonic_velocity_interface)
        )
        R = R.at[momentum_index_y].set(
            fast_mode_weighting * velocity_y_interface
            + slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        R = R.at[momentum_index_z].set(
            fast_mode_weighting * velocity_z_interface
            + slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        R = R.at[magnetic_index_y].set(
            sound_speed_interface * slow_mode_weighting * bt_normalized_y / sqrt_rho
        )
        R = R.at[magnetic_index_z].set(
            sound_speed_interface * slow_mode_weighting * bt_normalized_z / sqrt_rho
        )
        R = R.at[energy_index].set(
            fast_mode_weighting
            * (
                fast_magnetosonic_velocity_interface**2
                - fast_magnetosonic_velocity_interface * velocity_x_interface
                + 0.5 * velocity_sq_interface
                - gam2 * sound_speed_sq_interface
            )
            + slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        R = jnp.where(~sound_speed_greater_alfven_speed, R * sgn_bt, R)
        return R

    def col_1():
        # Column 2 (alfven -)
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(0.0)
        R = R.at[momentum_index_x].set(0.0)
        R = R.at[momentum_index_y].set(-bt_normalized_z)
        R = R.at[momentum_index_z].set(bt_normalized_y)
        R = R.at[magnetic_index_y].set(-bt_normalized_z * sgn_bx / sqrt_rho)
        R = R.at[magnetic_index_z].set(bt_normalized_y * sgn_bx / sqrt_rho)
        R = R.at[energy_index].set(
            bt_normalized_y * velocity_z_interface
            - bt_normalized_z * velocity_y_interface
        )
        return R

    def col_2():
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(slow_mode_weighting)
        R = R.at[momentum_index_x].set(
            slow_mode_weighting
            * (velocity_x_interface - slow_magnetosonic_velocity_interface)
        )
        R = R.at[momentum_index_y].set(
            slow_mode_weighting * velocity_y_interface
            - fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        R = R.at[momentum_index_z].set(
            slow_mode_weighting * velocity_z_interface
            - fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        R = R.at[magnetic_index_y].set(
            -sound_speed_interface * fast_mode_weighting * bt_normalized_y / sqrt_rho
        )
        R = R.at[magnetic_index_z].set(
            -sound_speed_interface * fast_mode_weighting * bt_normalized_z / sqrt_rho
        )
        R = R.at[energy_index].set(
            slow_mode_weighting
            * (
                slow_magnetosonic_velocity_interface**2
                - slow_magnetosonic_velocity_interface * velocity_x_interface
                + 0.5 * velocity_sq_interface
                - gam2 * sound_speed_sq_interface
            )
            - fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        R = jnp.where(sound_speed_greater_alfven_speed, R * sgn_bt, R)
        return R

    def col_3():
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(1.0)
        R = R.at[momentum_index_x].set(velocity_x_interface)
        R = R.at[momentum_index_y].set(velocity_y_interface)
        R = R.at[momentum_index_z].set(velocity_z_interface)
        R = R.at[magnetic_index_y].set(0.0)
        R = R.at[magnetic_index_z].set(0.0)
        R = R.at[energy_index].set(0.5 * velocity_sq_interface)
        return R

    def col_4():
        # Column 5 (slow +)
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(slow_mode_weighting)
        R = R.at[momentum_index_x].set(
            slow_mode_weighting
            * (velocity_x_interface + slow_magnetosonic_velocity_interface)
        )
        R = R.at[momentum_index_y].set(
            slow_mode_weighting * velocity_y_interface
            + fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        R = R.at[momentum_index_z].set(
            slow_mode_weighting * velocity_z_interface
            + fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        R = R.at[magnetic_index_y].set(
            -sound_speed_interface * fast_mode_weighting * bt_normalized_y / sqrt_rho
        )
        R = R.at[magnetic_index_z].set(
            -sound_speed_interface * fast_mode_weighting * bt_normalized_z / sqrt_rho
        )
        R = R.at[energy_index].set(
            slow_mode_weighting
            * (
                slow_magnetosonic_velocity_interface**2
                + slow_magnetosonic_velocity_interface * velocity_x_interface
                + 0.5 * velocity_sq_interface
                - gam2 * sound_speed_sq_interface
            )
            + fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        R = jnp.where(sound_speed_greater_alfven_speed, R * sgn_bt, R)
        return R

    def col_5():
        # Column 6 (alfven +)
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(0.0)
        R = R.at[momentum_index_x].set(0.0)
        R = R.at[momentum_index_y].set(-bt_normalized_z)
        R = R.at[momentum_index_z].set(bt_normalized_y)
        R = R.at[magnetic_index_y].set(bt_normalized_z * sgn_bx / sqrt_rho)
        R = R.at[magnetic_index_z].set(-bt_normalized_y * sgn_bx / sqrt_rho)
        R = R.at[energy_index].set(
            bt_normalized_y * velocity_z_interface
            - bt_normalized_z * velocity_y_interface
        )
        return R

    def col_6():
        # Column 7 (fast +)
        R = jnp.zeros_like(conserved_state)
        R = R.at[density_index].set(fast_mode_weighting)
        R = R.at[momentum_index_x].set(
            fast_mode_weighting
            * (velocity_x_interface + fast_magnetosonic_velocity_interface)
        )
        R = R.at[momentum_index_y].set(
            fast_mode_weighting * velocity_y_interface
            - slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        R = R.at[momentum_index_z].set(
            fast_mode_weighting * velocity_z_interface
            - slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        R = R.at[magnetic_index_y].set(
            sound_speed_interface * slow_mode_weighting * bt_normalized_y / sqrt_rho
        )
        R = R.at[magnetic_index_z].set(
            sound_speed_interface * slow_mode_weighting * bt_normalized_z / sqrt_rho
        )
        R = R.at[energy_index].set(
            fast_mode_weighting
            * (
                fast_magnetosonic_velocity_interface**2
                + fast_magnetosonic_velocity_interface * velocity_x_interface
                + 0.5 * velocity_sq_interface
                - gam2 * sound_speed_sq_interface
            )
            - slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        R = jnp.where(~sound_speed_greater_alfven_speed, R * sgn_bt, R)
        return R

    R = jax.lax.switch(col, [col_0, col_1, col_2, col_3, col_4, col_5, col_6])

    return R


@partial(jax.jit, static_argnames=["registered_variables"])
def _eigen_L_row(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    registered_variables: RegisteredVariables,
    row: int,
):
    (
        rho_interface,
        sqrt_rho,
        velocity_x_interface,
        velocity_y_interface,
        velocity_z_interface,
        velocity_sq_interface,
        magnetic_x_interface,
        magnetic_y_interface,
        magnetic_z_interface,
        bt_normalized_y,
        bt_normalized_z,
        sgn_bx,
        sgn_bt,
        sound_speed_interface,
        sound_speed_sq_interface,
        sound_speed_sq_inverse,
        sound_speed_greater_alfven_speed,
        fast_magnetosonic_velocity_interface,
        alfven_velocity_interface,
        slow_magnetosonic_velocity_interface,
        fast_mode_weighting,
        slow_mode_weighting,
        gam0,
        gam1,
        gam2,
    ) = _eigenvector_building_blocks(
        conserved_state,
        gamma,
        rhomin,
        pgmin,
        registered_variables,
    )

    # shorter names for registry indices
    density_index = registered_variables.density_index
    momentum_index_x = registered_variables.momentum_index.x
    momentum_index_y = registered_variables.momentum_index.y
    momentum_index_z = registered_variables.momentum_index.z
    magnetic_index_y = registered_variables.magnetic_index.y
    magnetic_index_z = registered_variables.magnetic_index.z
    energy_index = registered_variables.energy_index

    def row_0():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            fast_mode_weighting
            * (
                gam1 * velocity_sq_interface
                + fast_magnetosonic_velocity_interface * velocity_x_interface
            )
            - slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        L = L.at[momentum_index_x].set(
            fast_mode_weighting
            * (gam0 * velocity_x_interface - fast_magnetosonic_velocity_interface)
        )
        L = L.at[momentum_index_y].set(
            gam0 * fast_mode_weighting * velocity_y_interface
            + slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        L = L.at[momentum_index_z].set(
            gam0 * fast_mode_weighting * velocity_z_interface
            + slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        L = L.at[magnetic_index_y].set(
            gam0 * fast_mode_weighting * magnetic_y_interface
            + sound_speed_interface * slow_mode_weighting * bt_normalized_y * sqrt_rho
        )
        L = L.at[magnetic_index_z].set(
            gam0 * fast_mode_weighting * magnetic_z_interface
            + sound_speed_interface * slow_mode_weighting * bt_normalized_z * sqrt_rho
        )
        L = L.at[energy_index].set(-gam0 * fast_mode_weighting)
        L = 0.5 * L * sound_speed_sq_inverse
        L = jnp.where(~sound_speed_greater_alfven_speed, L * sgn_bt, L)
        return L

    def row_1():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            bt_normalized_z * velocity_y_interface
            - bt_normalized_y * velocity_z_interface
        )
        L = L.at[momentum_index_x].set(0.0)
        L = L.at[momentum_index_y].set(-bt_normalized_z)
        L = L.at[momentum_index_z].set(bt_normalized_y)
        L = L.at[magnetic_index_y].set(-bt_normalized_z * sgn_bx * sqrt_rho)
        L = L.at[magnetic_index_z].set(bt_normalized_y * sgn_bx * sqrt_rho)
        L = L.at[energy_index].set(0.0)
        L = 0.5 * L
        return L

    def row_2():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            slow_mode_weighting
            * (
                gam1 * velocity_sq_interface
                + slow_magnetosonic_velocity_interface * velocity_x_interface
            )
            + fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        L = L.at[momentum_index_x].set(
            gam0 * slow_mode_weighting * velocity_x_interface
            - slow_mode_weighting * slow_magnetosonic_velocity_interface
        )
        L = L.at[momentum_index_y].set(
            gam0 * slow_mode_weighting * velocity_y_interface
            - fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        L = L.at[momentum_index_z].set(
            gam0 * slow_mode_weighting * velocity_z_interface
            - fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        L = L.at[magnetic_index_y].set(
            gam0 * slow_mode_weighting * magnetic_y_interface
            - sound_speed_interface * fast_mode_weighting * bt_normalized_y * sqrt_rho
        )
        L = L.at[magnetic_index_z].set(
            gam0 * slow_mode_weighting * magnetic_z_interface
            - sound_speed_interface * fast_mode_weighting * bt_normalized_z * sqrt_rho
        )
        L = L.at[energy_index].set(-gam0 * slow_mode_weighting)
        L = 0.5 * L * sound_speed_sq_inverse
        L = jnp.where(sound_speed_greater_alfven_speed, L * sgn_bt, L)
        return L

    def row_3():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            -sound_speed_sq_interface / gam0 - 0.5 * velocity_sq_interface
        )
        L = L.at[momentum_index_x].set(velocity_x_interface)
        L = L.at[momentum_index_y].set(velocity_y_interface)
        L = L.at[momentum_index_z].set(velocity_z_interface)
        L = L.at[magnetic_index_y].set(magnetic_y_interface)
        L = L.at[magnetic_index_z].set(magnetic_z_interface)
        L = L.at[energy_index].set(-1.0)
        L = -gam0 * L * sound_speed_sq_inverse
        return L

    def row_4():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            slow_mode_weighting
            * (
                gam1 * velocity_sq_interface
                - slow_magnetosonic_velocity_interface * velocity_x_interface
            )
            - fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        L = L.at[momentum_index_x].set(
            slow_mode_weighting
            * (gam0 * velocity_x_interface + slow_magnetosonic_velocity_interface)
        )
        L = L.at[momentum_index_y].set(
            gam0 * slow_mode_weighting * velocity_y_interface
            + fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        L = L.at[momentum_index_z].set(
            gam0 * slow_mode_weighting * velocity_z_interface
            + fast_mode_weighting
            * fast_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        L = L.at[magnetic_index_y].set(
            gam0 * slow_mode_weighting * magnetic_y_interface
            - sound_speed_interface * fast_mode_weighting * bt_normalized_y * sqrt_rho
        )
        L = L.at[magnetic_index_z].set(
            gam0 * slow_mode_weighting * magnetic_z_interface
            - sound_speed_interface * fast_mode_weighting * bt_normalized_z * sqrt_rho
        )
        L = L.at[energy_index].set(-gam0 * slow_mode_weighting)
        L = 0.5 * L * sound_speed_sq_inverse
        L = jnp.where(sound_speed_greater_alfven_speed, L * sgn_bt, L)
        return L

    def row_5():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            bt_normalized_z * velocity_y_interface
            - bt_normalized_y * velocity_z_interface
        )
        L = L.at[momentum_index_x].set(0.0)
        L = L.at[momentum_index_y].set(-bt_normalized_z)
        L = L.at[momentum_index_z].set(bt_normalized_y)
        L = L.at[magnetic_index_y].set(bt_normalized_z * sgn_bx * sqrt_rho)
        L = L.at[magnetic_index_z].set(-bt_normalized_y * sgn_bx * sqrt_rho)
        L = L.at[energy_index].set(0.0)
        L = 0.5 * L
        return L

    def row_6():
        L = jnp.zeros_like(conserved_state)
        L = L.at[density_index].set(
            fast_mode_weighting
            * (
                gam1 * velocity_sq_interface
                - fast_magnetosonic_velocity_interface * velocity_x_interface
            )
            + slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * (
                bt_normalized_y * velocity_y_interface
                + bt_normalized_z * velocity_z_interface
            )
            * sgn_bx
        )
        L = L.at[momentum_index_x].set(
            fast_mode_weighting
            * (gam0 * velocity_x_interface + fast_magnetosonic_velocity_interface)
        )
        L = L.at[momentum_index_y].set(
            gam0 * fast_mode_weighting * velocity_y_interface
            - slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_y
            * sgn_bx
        )
        L = L.at[momentum_index_z].set(
            gam0 * fast_mode_weighting * velocity_z_interface
            - slow_mode_weighting
            * slow_magnetosonic_velocity_interface
            * bt_normalized_z
            * sgn_bx
        )
        L = L.at[magnetic_index_y].set(
            gam0 * fast_mode_weighting * magnetic_y_interface
            + sound_speed_interface * slow_mode_weighting * bt_normalized_y * sqrt_rho
        )
        L = L.at[magnetic_index_z].set(
            gam0 * fast_mode_weighting * magnetic_z_interface
            + sound_speed_interface * slow_mode_weighting * bt_normalized_z * sqrt_rho
        )
        L = L.at[energy_index].set(-gam0 * fast_mode_weighting)
        L = 0.5 * L * sound_speed_sq_inverse
        L = jnp.where(~sound_speed_greater_alfven_speed, L * sgn_bt, L)
        return L

    L = jax.lax.switch(row, [row_0, row_1, row_2, row_3, row_4, row_5, row_6])

    return L


@partial(jax.jit, static_argnames=["registered_variables"])
def _eigen_all_lambdas(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    registered_variables: RegisteredVariables,
):
    (
        velocity_x,
        fast_magnetosonic_velocity,
        alfven_velocity,
        slow_magnetosonic_velocity,
    ) = _eigenvalue_building_blocks(
        conserved_state,
        gamma,
        rhomin,
        pgmin,
        registered_variables,
    )

    return jnp.stack(
        [
            velocity_x - fast_magnetosonic_velocity,
            velocity_x - alfven_velocity,
            velocity_x - slow_magnetosonic_velocity,
            velocity_x,
            velocity_x + slow_magnetosonic_velocity,
            velocity_x + alfven_velocity,
            velocity_x + fast_magnetosonic_velocity,
        ],
        axis=0,
    )


def _eigen_lambdas(
    conserved_state,
    rhomin: Union[float, jnp.ndarray],
    pgmin: Union[float, jnp.ndarray],
    gamma: Union[float, jnp.ndarray],
    registered_variables: RegisteredVariables,
    mode: int,
):
    (
        velocity_x,
        fast_magnetosonic_velocity,
        alfven_velocity,
        slow_magnetosonic_velocity,
    ) = _eigenvalue_building_blocks(
        conserved_state,
        gamma,
        rhomin,
        pgmin,
        registered_variables,
    )

    def mode_0():
        return velocity_x - fast_magnetosonic_velocity

    def mode_1():
        return velocity_x - alfven_velocity

    def mode_2():
        return velocity_x - slow_magnetosonic_velocity

    def mode_3():
        return velocity_x

    def mode_4():
        return velocity_x + slow_magnetosonic_velocity

    def mode_5():
        return velocity_x + alfven_velocity

    def mode_6():
        return velocity_x + fast_magnetosonic_velocity

    return jax.lax.switch(
        mode, [mode_0, mode_1, mode_2, mode_3, mode_4, mode_5, mode_6]
    )



from types import NoneType
from typing import NamedTuple, Union

from astronomix.option_classes.simulation_config import STATE_TYPE


class StateStruct(NamedTuple):
    """
    Struct for the simulation state.
    """

    #: The fluid state.
    primitive_state: Union[STATE_TYPE, NoneType] = None

    # here you might add more fields like the 
    # positions of star particles, etc.



from types import NoneType
from typing import NamedTuple, Union

from jaxtyping import PyTree

class NeuralNetForceConfig(NamedTuple):
    neural_net_force: bool = False
    network_static: Union[PyTree, NoneType] = None

class NeuralNetForceParams(NamedTuple):
    network_params: Union[PyTree, NoneType] = None


from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import FIELD_TYPE, FINITE_DIFFERENCE, STATE_TYPE, SimulationConfig

from typing import Union

import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import jaxtyped


from functools import partial
from types import NoneType


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables", "config", "sharding"])
def construct_primitive_state(
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    density: FIELD_TYPE,
    velocity_x: Union[FIELD_TYPE, NoneType] = None,
    velocity_y: Union[FIELD_TYPE, NoneType] = None,
    velocity_z: Union[FIELD_TYPE, NoneType] = None,
    magnetic_field_x: Union[FIELD_TYPE, NoneType] = None,
    magnetic_field_y: Union[FIELD_TYPE, NoneType] = None,
    magnetic_field_z: Union[FIELD_TYPE, NoneType] = None,
    interface_magnetic_field_x: Union[FIELD_TYPE, NoneType] = None,
    interface_magnetic_field_y: Union[FIELD_TYPE, NoneType] = None,
    interface_magnetic_field_z: Union[FIELD_TYPE, NoneType] = None,
    gas_pressure: Union[FIELD_TYPE, NoneType] = None,
    cosmic_ray_pressure: Union[FIELD_TYPE, NoneType] = None,
    sharding=None,
) -> STATE_TYPE:
    """Stack the primitive variables into the state array.

    IN 1D SET ONLY THE XCOMPONENTS, in 2D SET X AND Y COMPONENTS,
    in 3D SET X, Y AND Z COMPONENTS

    Args:
        config: The simulation configuration.
        registered_variables: The indices of the variables in the state array.
        density: The density of the fluid.
        velocity_x: The x-component of the velocity of the fluid.
        velocity_y: The y-component of the velocity of the fluid.
        velocity_z: The z-component of the velocity of the fluid.
        magnetic_field_x: The x-component of the magnetic field in B / sqrt(\mu_0).
        magnetic_field_y: The y-component of the magnetic field in B / sqrt(\mu_0).
        magnetic_field_z: The z-component of the magnetic field in B / sqrt(\mu_0).
        gas_pressure: The thermal pressure of the fluid.
        cosmic_ray_pressure: The cosmic ray pressure of the fluid.

    Returns:
        The state array.
    """
    if sharding is not None:
        state = jax.lax.with_sharding_constraint(
            jnp.zeros((registered_variables.num_vars, *density.shape)), sharding
        )
    else:
        state = jnp.zeros((registered_variables.num_vars, *density.shape))

    state = state.at[registered_variables.density_index].set(density)

    if config.dimensionality == 1:
        state = state.at[registered_variables.velocity_index].set(velocity_x)
    elif config.dimensionality == 2:
        state = state.at[registered_variables.velocity_index.x].set(velocity_x)
        state = state.at[registered_variables.velocity_index.y].set(velocity_y)
    elif config.dimensionality == 3:
        state = state.at[registered_variables.velocity_index.x].set(velocity_x)
        state = state.at[registered_variables.velocity_index.y].set(velocity_y)
        state = state.at[registered_variables.velocity_index.z].set(velocity_z)

    if config.mhd:
        if config.dimensionality >= 2:
            if magnetic_field_x is not None:
                state = state.at[registered_variables.magnetic_index.x].set(
                    magnetic_field_x
                )
            if magnetic_field_y is not None:
                state = state.at[registered_variables.magnetic_index.y].set(
                    magnetic_field_y
                )
            if magnetic_field_z is not None:
                state = state.at[registered_variables.magnetic_index.z].set(
                    magnetic_field_z
            )

        if config.solver_mode == FINITE_DIFFERENCE:
            if interface_magnetic_field_x is not None:
                state = state.at[
                    registered_variables.interface_magnetic_field_index.x
                ].set(interface_magnetic_field_x)
            if interface_magnetic_field_y is not None:
                state = state.at[
                    registered_variables.interface_magnetic_field_index.y
                ].set(interface_magnetic_field_y)
            if interface_magnetic_field_z is not None:
                state = state.at[
                    registered_variables.interface_magnetic_field_index.z
                ].set(interface_magnetic_field_z)
    
    state = state.at[registered_variables.pressure_index].set(gas_pressure)

    if registered_variables.cosmic_ray_n_active:
        # TODO: get from params
        gamma_cr = 4 / 3

        state = state.at[registered_variables.pressure_index].set(
            gas_pressure + cosmic_ray_pressure
        )
        state = state.at[registered_variables.cosmic_ray_n_index].set(
            cosmic_ray_pressure ** (1 / gamma_cr)
        )

    return state


# general
from functools import partial
import jax
import jax.numpy as jnp

# typing
from beartype import beartype as typechecker
from typing import Union
from jaxtyping import Array, Float, jaxtyped

# jf1uid constants
from astronomix.option_classes.simulation_config import FIELD_TYPE, PERIODIC_BOUNDARY

# astronomix classes
from astronomix.option_classes.simulation_config import SimulationConfig

# fft
from jax.numpy.fft import fftn, ifftn


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["grid_spacing", "config"])
def _compute_gravitational_potential(
    gas_density: FIELD_TYPE,
    grid_spacing: float,
    config: SimulationConfig,
    G: Union[float, Float[Array, ""]] = 1.0,
) -> FIELD_TYPE:
    """
    Compute the gravitational potential using FFT to solve Poisson's equation for
    periodic and open boundaries (via the Hockney & Eastwood method).

    Args:
        gas_density: The gas density field.
        grid_spacing: The grid spacing.
        config: The simulation configuration.
        G: The gravitational constant.

    Returns:
        The gravitational potential.

    """

    # TODO: remove ghost cells in this computations (?)

    dimensionality = config.dimensionality
    num_cells = gas_density.shape[0]

    # we only use outflow if not all boundaries are periodic
    # SO THERES EITHER ALL PERIODIC OR NONE
    # TODO: improve

    non_periodic_boundaries = False

    if dimensionality == 1:
        if not (
            config.boundary_settings.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.right_boundary == PERIODIC_BOUNDARY
        ):
            non_periodic_boundaries = True
    elif dimensionality == 2:
        if not (
            config.boundary_settings.x.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.x.right_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.y.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.y.right_boundary == PERIODIC_BOUNDARY
        ):
            non_periodic_boundaries = True
    elif dimensionality == 3:
        if not (
            config.boundary_settings.x.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.x.right_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.y.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.y.right_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.z.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.z.right_boundary == PERIODIC_BOUNDARY
        ):
            non_periodic_boundaries = True

    # if periodic boundaries
    if not non_periodic_boundaries:
        # jeans swindle (?)
        gas_density = gas_density - jnp.mean(gas_density)

    if not non_periodic_boundaries:
        # -----------------------------
        # Periodic boundaries version
        # -----------------------------
        # Compute FFT of the density field.
        density_k = fftn(gas_density)

        # Build the k–vector (note that fftfreq returns cycles/unit; multiply by 2π).
        k_base = jnp.fft.fftfreq(num_cells, d=grid_spacing) * 2 * jnp.pi

        if dimensionality == 1:
            k = k_base  # 1D case.
            k_squared = k**2
        elif dimensionality == 2:
            kx, ky = jnp.meshgrid(k_base, k_base, indexing="ij")
            k_squared = kx**2 + ky**2
        elif dimensionality == 3:
            kx, ky, kz = jnp.meshgrid(k_base, k_base, k_base, indexing="ij")
            k_squared = kx**2 + ky**2 + kz**2

        # Avoid division by zero (k = 0 mode).
        k_squared = jnp.where(k_squared == 0, 1e-12, k_squared)
        greens_function = jnp.where(
            k_squared > 1e-12, -4 * jnp.pi * G / k_squared, -1 / (4 * jnp.pi)
        )

        # Multiply in Fourier space and invert.
        potential_k = greens_function * density_k
        gravitational_potential = jnp.real(ifftn(potential_k))

        return gravitational_potential

    else:
        # ----------------------------------------------------
        # Open boundaries version via Hockney & Eastwood method
        # ----------------------------------------------------
        #
        # (a) Extend the domain to twice the size in each dimension.
        original_shape = gas_density.shape
        extended_shape = tuple(2 * s for s in original_shape)

        # Embed the original density in the (0,...,0) corner of the extended grid.
        extended_density = jnp.zeros(extended_shape, dtype=gas_density.dtype)
        slices = tuple(slice(0, s) for s in original_shape)
        extended_density = extended_density.at[slices].set(gas_density)

        # (b) Construct the Green's function on the extended grid.
        #
        # The Hockney–Eastwood prescription is to compute, for each dimension,
        #     pos = [0, 1, 2, ..., n-1, 2n - n, ..., 1] * grid_spacing,
        # which is equivalent to:
        #
        #    pos = np.arange(2*n);  pos = where(pos < n, pos, 2*n - pos)
        #
        # This yields the “minimum–image” distances from the source placed at the origin.
        grids = []
        for s in original_shape:
            n = s
            extended_n = 2 * n
            pos = jnp.arange(extended_n)
            pos = jnp.where(pos < n, pos, 2 * n - pos)
            pos = pos * grid_spacing
            grids.append(pos)

        # Now create the distance array r on the extended grid.
        if dimensionality == 1:
            r = grids[0]  # Already nonnegative.
        elif dimensionality == 2:
            x, y = jnp.meshgrid(grids[0], grids[1], indexing="ij")
            r = jnp.sqrt(x**2 + y**2)
        elif dimensionality == 3:
            x, y, z = jnp.meshgrid(grids[0], grids[1], grids[2], indexing="ij")
            r = jnp.sqrt(x**2 + y**2 + z**2)

        # Replace any zero distance with grid_spacing (to avoid singularity).
        r_safe = jnp.where(r == 0, grid_spacing, r)

        # (c) Define the isolated (open–boundary) Green's function.
        if dimensionality == 1:
            # For 1D (solving φ'' = 4πGδ(x)), the solution is φ = -2πG |x|
            kernel = -2 * jnp.pi * G * r
        elif dimensionality == 2:
            # In 2D, φ = -2G log(r)  (up to an additive constant).
            kernel = -2 * G * jnp.log(r_safe)
        elif dimensionality == 3:
            # In 3D, the isolated potential is φ = -G/r.
            kernel = -G / r_safe

        # (d) FFT–convolve: Multiply the FFTs of the extended density and the Green's function.
        density_k_ext = fftn(extended_density)
        kernel_k_ext = fftn(kernel)
        potential_ext = jnp.real(ifftn(density_k_ext * kernel_k_ext))

        # (e) Extract the portion of the potential corresponding to the original grid.
        gravitational_potential = potential_ext[slices]
        return gravitational_potential / num_cells**dimensionality



import jax
import jax.numpy as jnp


def _cosine_taper(k, k0, k1):
    """Returns taper factor in [0,1]: 1 for k<=k0, smooth down to 0 at k>=k1."""
    # If k1 == k0, use a hard cutoff
    small = 1e-12
    if k1 <= k0 + small:
        return jnp.where(k <= k0, 1.0, 0.0)
    t = (k - k0) / (k1 - k0)
    t = jnp.clip(t, 0.0, 1.0)
    return jnp.where(k <= k0, 1.0, 0.5 * (1.0 + jnp.cos(jnp.pi * t)))


# k_max <= int(0.7 * Ndim/2)


def create_turb_field(
    Ndim,
    A0,
    slope,
    kmin,
    kmax,
    key,
    sharding=None,
    kroll_frac=0.85,
    zero_mean=True,
):
    """
    Generate a real Gaussian random field with target amplitude scaling.

    This version is optimized for memory by avoiding explicit meshgrid/indices
    arrays and using broadcasting and efficient array manipulations instead.
    """
    # Build integer wavenumbers [-N/2..N/2-1]
    k1d = jnp.fft.fftfreq(Ndim, d=1.0) * Ndim

    # --- Memory Optimization 1: Broadcasting instead of meshgrid ---
    # Calculate k-space magnitude without creating full kx, ky, kz arrays.
    # JAX will broadcast the 1D arrays to 3D during the operation.
    k3d_sq = (
        k1d.reshape(Ndim, 1, 1) ** 2
        + k1d.reshape(1, Ndim, 1) ** 2
        + k1d.reshape(1, 1, Ndim) ** 2
    )

    # Shard the first large array created. Subsequent element-wise ops
    # will inherit the sharding.
    if sharding is not None:
        k3d_sq = jax.device_put(k3d_sq, sharding)

    k3d = jnp.sqrt(k3d_sq)

    # Optional roll-off: start rolling at kroll_frac * kmax and finish at kmax
    k_roll_start = kroll_frac * kmax
    taper = _cosine_taper(k3d, k_roll_start, kmax)

    # Mask inside k-band where we want power
    band_mask = (k3d >= kmin) & (k3d <= kmax)

    # Set amplitude, avoiding division by zero at k=0
    k_safe = jnp.where(k3d == 0.0, 1.0, k3d)
    amplitude = A0 * (k_safe**slope)
    if zero_mean:
        amplitude = amplitude.at[0, 0, 0].set(0.0)

    # Apply band + taper
    amplitude = amplitude * band_mask * taper

    # Sample complex Fourier coefficients with variance ~ amplitude^2
    sigma = amplitude / jnp.sqrt(2.0)
    subkeys = jax.random.split(key, 2)
    re = jax.random.normal(subkeys[0], shape=k3d.shape) * sigma
    im = jax.random.normal(subkeys[1], shape=k3d.shape) * sigma
    F = re + 1j * im

    # --- Memory Optimization 2: Enforce Hermitian symmetry without indices ---
    # The operation F_conj_flipped[-k] is equivalent to flipping all axes and
    # rolling by one element due to the fftfreq convention. This is much
    # cheaper than a gather operation with explicit index arrays.
    F_conj_flipped = jnp.roll(
        jnp.flip(jnp.conj(F), axis=(0, 1, 2)), shift=1, axis=(0, 1, 2)
    )
    F_sym = 0.5 * (F + F_conj_flipped)

    # --- Memory Optimization 3: Find self-conjugate modes without indices ---
    # Self-conjugate modes are where k_i = -k_i (mod N), which occurs at
    # indices 0 and N/2 (for even N). We can build this mask via broadcasting.
    idx_1d = jnp.arange(Ndim)
    sc_1d = idx_1d == ((-idx_1d) % Ndim)
    self_conj_mask = (
        sc_1d.reshape(Ndim, 1, 1)
        & sc_1d.reshape(1, Ndim, 1)
        & sc_1d.reshape(1, 1, Ndim)
    )

    # Force imaginary part to be zero at self-conjugate locations
    F_sym = jnp.where(self_conj_mask, jnp.real(F_sym).astype(F_sym.dtype), F_sym)

    # Optional: explicitly ensure DC is real zero (if zero_mean True)
    if zero_mean:
        F_sym = F_sym.at[0, 0, 0].set(0.0 + 0.0j)

    # Inverse transform back to real space
    rfield_complex = jnp.fft.ifftn(F_sym)
    rfield = jnp.real(rfield_complex)

    return rfield



from functools import partial
import jax.numpy as jnp
import jax

from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union

from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import (
    total_energy_from_primitives_with_crs,
)
from astronomix._fluid_equations._equations import (
    get_absolute_velocity,
    total_energy_from_primitives,
    # total_pressure_from_conserved_mhd,
)
from astronomix.variable_registry.registered_variables import AxisInfo, RegisteredVariables
from astronomix.option_classes.simulation_config import STATE_TYPE, SimulationConfig


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["config", "registered_variables", "flux_direction_index"]
)
def _euler_flux(
    primitive_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    flux_direction_index: int,
) -> STATE_TYPE:
    """Compute the Euler fluxes for the given primitive states.

    Args:
        primitive_state: The primitive state of the fluid on all cells.
        gamma: The adiabatic index of the fluid.
        registered_variables: The registered variables.
        flux_direction_index: The index of the velocity component in the flux direction of interest.

    Returns:
        The Euler fluxes for the given primitive states.
    """
    rho = primitive_state[registered_variables.density_index]
    p = primitive_state[registered_variables.pressure_index]

    # start with a copy of the primitive states
    flux_vector = (
        primitive_state # .copy()
    )  # copy should not be necessary (no side effects)

    # calculate the total energy
    utotal = get_absolute_velocity(primitive_state, config, registered_variables)

    if registered_variables.cosmic_ray_n_active:
        E = total_energy_from_primitives_with_crs(primitive_state, registered_variables)
    else:
        E = total_energy_from_primitives(rho, utotal, p, gamma)

    # add the total energy to the pressure_index of the flux vector
    flux_vector = flux_vector.at[registered_variables.pressure_index].add(E)

    # scale the velocity components with the density
    if config.dimensionality == 1:
        flux_vector = flux_vector.at[registered_variables.velocity_index].set(
            primitive_state[registered_variables.velocity_index] * rho
        )
    elif config.dimensionality == 2:
        flux_vector = flux_vector.at[registered_variables.velocity_index.x].set(
            primitive_state[registered_variables.velocity_index.x] * rho
        )
        flux_vector = flux_vector.at[registered_variables.velocity_index.y].set(
            primitive_state[registered_variables.velocity_index.y] * rho
        )
    elif config.dimensionality == 3:
        flux_vector = flux_vector.at[registered_variables.velocity_index.x].set(
            primitive_state[registered_variables.velocity_index.x] * rho
        )
        flux_vector = flux_vector.at[registered_variables.velocity_index.y].set(
            primitive_state[registered_variables.velocity_index.y] * rho
        )
        flux_vector = flux_vector.at[registered_variables.velocity_index.z].set(
            primitive_state[registered_variables.velocity_index.z] * rho
        )

    # multiply the whole vector with the velocity component in the flux direction
    # primitive_state[flux_direction_index] only works because of how the registry is
    # currently ordered: TODO: generalize
    flux_vector = primitive_state[flux_direction_index] * flux_vector

    # add the pressure to the velocity component in the flux direction
    flux_vector = flux_vector.at[flux_direction_index].add(p)

    return flux_vector

# # TODO: improve readability
# @partial(
#     jax.jit, static_argnames=["config", "registered_variables", "flux_axis"]
# )
# def _mhd_flux(
#     conserved_state: STATE_TYPE,
#     gamma: Union[float, Float[Array, ""]],
#     config: SimulationConfig,
#     registered_variables: RegisteredVariables,
#     flux_axis: AxisInfo,
# ) -> STATE_TYPE:
#     """
#     Compute the MHD fluxes, for now only in 3D.

#     NOTE: We work on the conserved state here, not the primitive state!
#     (Because I'm currently writing this for the finite difference MHD solver, 
#     where we handle conserved variables.)

#     Conserved state:
#         [ ρ,

#           ρ v_x,
#           ρ v_y,
#           ρ v_z,

#           E,

#           B_x,
#           B_y,
#           B_z,

#         ]

#     The flux in x-direction is given by:

#         F = [
#             ρ v_x,                         // density_index

#             ρ v_x^2 + P_tot - B_x^2,       // velocity_index.x
#             ρ v_x v_y - B_x B_y,           // velocity_index.y
#             ρ v_x v_z - B_x B_z,           // velocity_index.z

#             (E + P_tot) v_x - (B · v) B_x  // energy_index

#             0,                             // magnetic_index.x
#             B_y v_x - B_x v_y,             // magnetic_index.y
#             B_z v_x - B_x v_z,             // magnetic_index.z

#         ]

#     The flux in y-direction is given by:
#         F = [
#             ρ v_y,                         // density_index

#             ρ v_y v_x - B_y B_x,           // velocity_index.x
#             ρ v_y^2 + P_tot - B_y^2,       // velocity_index.y
#             ρ v_y v_z - B_y B_z,           // velocity_index.z

#             (E + P_tot) v_y - (B · v) B_y  // energy_index

#             B_x v_y - B_y v_x,             // magnetic_index.x
#             0,                             // magnetic_index.y
#             B_z v_y - B_y v_z,             // magnetic_index.z

#         ]

#     The flux in z-direction is given by:
#         F = [
#             ρ v_z,                         // density_index

#             ρ v_z v_x - B_z B_x,           // velocity_index.x
#             ρ v_z v_y - B_z B_y,           // velocity_index.y
#             ρ v_z^2 + P_tot - B_z^2,       // velocity_index.z

#             (E + P_tot) v_z - (B · v) B_z  // energy_index

#             B_x v_z - B_z v_x,             // magnetic_index.x
#             B_y v_z - B_z v_y,             // magnetic_index.y
#             0,                             // magnetic_index.z
#         ]

#     where P_tot = p + 0.5 * |B|^2 is the total pressure.

#     """

#     total_pressure = total_pressure_from_conserved_mhd(
#         conserved_state, gamma, registered_variables
#     )

#     # we can get from the conserved state to the flux by
#     # - add the total pressure to the pressure_index
#     # - multipy everything with conserved_state[axis_info.velocity_index] / density
#     # - add the pressure term to the axis_info.velocity_index
#     # - subtract the the magnetic terms * axis_info.magnetic_index from the velocity components
#     # - subtract the velocity terms * axis_info.magnetic_index from the magnetic components
#     #  - subtract the (B · v) B_axis from the energy component
#     # - set the field at axis_info.magnetic_index to zero

#     flux_vector = conserved_state.at[registered_variables.pressure_index].add(
#         total_pressure
#     )

#     v_axis = conserved_state[flux_axis.velocity_index] / conserved_state[
#         registered_variables.density_index
#     ]

#     flux_vector = v_axis * flux_vector

#     flux_vector = flux_vector.at[flux_axis.velocity_index].add(
#         total_pressure
#     )

#     flux_vector = flux_vector.at[
#         registered_variables.velocity_index.x:registered_variables.velocity_index.z + 1
#     ].add(
#         -conserved_state[registered_variables.magnetic_index.x:registered_variables.magnetic_index.z + 1]
#         * conserved_state[flux_axis.magnetic_index]
#     )

#     flux_vector = flux_vector.at[
#         registered_variables.magnetic_index.x:registered_variables.magnetic_index.z + 1
#     ].add(
#         -conserved_state[
#             registered_variables.velocity_index.x:registered_variables.velocity_index.z + 1
#         ] / conserved_state[registered_variables.density_index]
#         * conserved_state[flux_axis.magnetic_index]
#     )

#     v_dot_B = (
#         conserved_state[registered_variables.magnetic_index.x]
#         * conserved_state[registered_variables.velocity_index.x]
#         + conserved_state[registered_variables.magnetic_index.y]
#         * conserved_state[registered_variables.velocity_index.y]
#         + conserved_state[registered_variables.magnetic_index.z]
#         * conserved_state[registered_variables.velocity_index.z]
#     ) / conserved_state[registered_variables.density_index]
#     flux_vector = flux_vector.at[registered_variables.pressure_index].add(
#         -v_dot_B * conserved_state[flux_axis.magnetic_index]
#     )

#     flux_vector = flux_vector.at[flux_axis.magnetic_index].set(0.0)

#     return flux_vector


"""
The registered_variables module tells the code where 
in the state array which field is stored. This is important 
for modularity and readability of the code.

When you want to add new variables to the state array, e.g.
densities for chemical species, you have to register them here.

NOTE: For finite volume MHD simulation, the magnetic field
is assumed to be stored in the last three indices of the state array
and for finite difference MHD the magnetic field at interfaces
is assumed to be stored in the three indices.
"""

from typing import NamedTuple

import jax.numpy as jnp

from jaxtyping import Array, Float, Int

from typing import Union

from astronomix.option_classes.simulation_config import FINITE_DIFFERENCE, XAXIS, YAXIS, ZAXIS, SimulationConfig


class StaticIntVector(NamedTuple):
    x: int
    y: int
    z: int

# =============================================================

"""
We perform simulations in multiple spatial dimensions. To an axis,
e.g. the x-axis, corresponds the axis in the array storing the state
(along which e.g. x varies) but also fields in the state related
to that axis, e.g. the x-velocity or the x-magnetic field component.

A common pattern will be a loop over the spatial dimensions where
we exactly need to know this information.
"""

class AxisInfo(NamedTuple):
    # corresponding axis in the array
    axis_in_array: int

    # corresponding field indices
    velocity_index: int
    magnetic_index: int

# =============================================================


class RegisteredVariables(NamedTuple):
    """
    The registered variables are the variables that are
    stored in the state array. The order of the variables
    in the state array is important and should be consistent
    throughout the code.
    """

    #: Axes information. For now only
    #: implemented for 3D simulations.
    axes_info: tuple[AxisInfo, ...] = ()

    #: Number of variables
    num_vars: int = 3

    # Baseline variables

    #: Density index
    density_index: int = 0

    #: Velocity index
    velocity_index: Union[int, StaticIntVector] = 1
    # in e.g. 3D, we have three velocity components, each with its own index

    #: Momentum density index, same as velocity index
    #: introduced for readability when dealing with
    #: the conserved state
    momentum_index: Union[int, StaticIntVector] = 1

    #: Magnetic field index
    magnetic_index: Union[int, StaticIntVector] = -1

    #: Magnetic field at interfaces index
    #: used in finite difference MHD constrained transport
    interface_magnetic_field_index: Union[int, StaticIntVector] = -1

    #: Pressure index
    pressure_index: int = 2

    #: Energy index, same as pressure index
    #: introduced for readability when dealing with
    #: the conserved state.
    energy_index: int = 2

    # Additional variables, these
    # have to be registered

    #: stellar wind density index
    wind_density_index: int = -1
    wind_density_active: bool = False

    #: simplified cosmic rays
    # in the simplest CR model witout CR diffusion,
    # streaming and no explicitly modeled magnetic field
    # n_CR = P_CR^(1/gamma_CR) is a conserved quantity.
    # This is the cosmic_ray_n, the index below points to.
    cosmic_ray_n_index: int = -1
    cosmic_ray_n_active: bool = False

    # here you can add more variables


def get_registered_variables(config: SimulationConfig) -> RegisteredVariables:
    """Get the registered variables for the simulation.

    Args:
        config: The simulation configuration.

    Returns:
        The registered variables.
    """

    registered_variables = RegisteredVariables()

    if config.dimensionality == 2:
        # we have two velocity components
        registered_variables = registered_variables._replace(
            num_vars=registered_variables.num_vars + 1
        )

        # update the velocity index
        registered_variables = registered_variables._replace(
            velocity_index=StaticIntVector(1, 2, -1)
        )

        # TODO: unified MHD approach in 1D/2D/3D
        # magnetic field index
        if config.mhd:
            # TODO: better indexing
            registered_variables = registered_variables._replace(pressure_index=3)
            registered_variables = registered_variables._replace(
                magnetic_index=StaticIntVector(4, 5, 6)
            )
            registered_variables = registered_variables._replace(
                num_vars=registered_variables.num_vars + 3
            )
        else:
            # update the pressure index
            registered_variables = registered_variables._replace(
                pressure_index=registered_variables.num_vars - 1
            )

    if config.dimensionality == 3:
        # we have three velocity components
        registered_variables = registered_variables._replace(
            num_vars=registered_variables.num_vars + 2
        )

        # update the velocity index to be an array
        registered_variables = registered_variables._replace(
            velocity_index=StaticIntVector(1, 2, 3)
        )

        # update the pressure index
        registered_variables = registered_variables._replace(
            pressure_index=registered_variables.num_vars - 1
        )

        # update the magnetic field index
        if config.mhd:
            registered_variables = registered_variables._replace(
                magnetic_index=StaticIntVector(5, 6, 7)
            )
            registered_variables = registered_variables._replace(
                num_vars=registered_variables.num_vars + 3
            )

    if config.mhd and config.solver_mode == FINITE_DIFFERENCE:
        # interface magnetic field indices
        registered_variables = registered_variables._replace(
            interface_magnetic_field_index=StaticIntVector(
                registered_variables.num_vars,
                registered_variables.num_vars + 1,
                registered_variables.num_vars + 2,
            )
        )
        registered_variables = registered_variables._replace(
            num_vars=registered_variables.num_vars + 3
        )

    if config.wind_config.trace_wind_density:
        registered_variables = registered_variables._replace(
            wind_density_index=registered_variables.num_vars
        )
        registered_variables = registered_variables._replace(
            num_vars=registered_variables.num_vars + 1
        )
        registered_variables = registered_variables._replace(wind_density_active=True)

    if config.cosmic_ray_config.cosmic_rays:
        registered_variables = registered_variables._replace(
            cosmic_ray_n_index=registered_variables.num_vars
        )
        registered_variables = registered_variables._replace(
            num_vars=registered_variables.num_vars + 1
        )
        registered_variables = registered_variables._replace(cosmic_ray_n_active=True)

    # axes info
    if config.dimensionality == 3 and config.mhd:
        axes_info = (
            AxisInfo(
                axis_in_array=XAXIS,
                velocity_index=registered_variables.velocity_index.x,
                magnetic_index=registered_variables.magnetic_index.x,
            ),
            AxisInfo(
                axis_in_array=YAXIS,
                velocity_index=registered_variables.velocity_index.y,
                magnetic_index=registered_variables.magnetic_index.y,
            ),
            AxisInfo(
                axis_in_array=ZAXIS,
                velocity_index=registered_variables.velocity_index.z,
                magnetic_index=registered_variables.magnetic_index.z,
            ),
        )
        registered_variables = registered_variables._replace(axes_info=axes_info)

    registered_variables = registered_variables._replace(momentum_index=registered_variables.velocity_index)
    registered_variables = registered_variables._replace(energy_index=registered_variables.pressure_index)

    # here you can register more variables

    return registered_variables



from types import NoneType
from typing import NamedTuple
from typing import Union
import jax.numpy as jnp
from jaxtyping import PyTree

SIMPLE_POWER_LAW = 1
PIECEWISE_POWER_LAW = 2
NEURAL_NET_COOLING = 3
NEURAL_NET_COOLING_WITH_DENSITY = 4

EXPLICIT_COOLING = 1
IMPLICIT_COOLING = 2


class SimplePowerLawParams(NamedTuple):
    factor: float = 1.0
    exponent: float = 1.0
    reference_temperature: float = 1e8

class PiecewisePowerLawParams(NamedTuple):
    log10_T_table: jnp.ndarray = jnp.array([])
    log10_Lambda_table: jnp.ndarray = jnp.array([])
    alpha_table: jnp.ndarray = jnp.array([])
    Y_table: jnp.ndarray = jnp.array([])
    reference_temperature: float = 1e8

class CoolingNetConfig(NamedTuple):
    network_static: Union[PyTree, NoneType] = None

class CoolingNetParams(NamedTuple):
    network_params: Union[PyTree, NoneType] = None

COOLING_CURVE_TYPE = Union[SimplePowerLawParams, PiecewisePowerLawParams, CoolingNetParams]

class CoolingCurveConfig(NamedTuple):
    cooling_curve_type: int = SIMPLE_POWER_LAW
    
    #: In case of neural the cooling the network architecture
    cooling_net_config: CoolingNetConfig = CoolingNetConfig()


class CoolingConfig(NamedTuple):
    cooling: bool = False
    cooling_method: int = IMPLICIT_COOLING
    cooling_curve_config: CoolingCurveConfig = CoolingCurveConfig()

class CoolingParams(NamedTuple):
    # NOTE: CURRENTLY ONLY POWER LAW COOLING
    hydrogen_mass_fraction: float = 0.76
    metal_mass_fraction: float = 0.02

    floor_temperature: float = 1e4

    cooling_curve_params: COOLING_CURVE_TYPE = SimplePowerLawParams()


"""
Constrained Transport (CT) implementation for MHD.
Based on HOW-MHD paper (Seo & Ryu 2023, 
see https://arxiv.org/abs/2304.04360).

Algorithm summary
-----------------

We carry interface magnetic fields

b_x at x-interfaces,
b_y at y-interfaces,
b_z at z-interfaces

through the simulation, updating them using the CT
algorithm such that (ignoring floating point errors) the
divergence of B remains zero. This is achieved by updating
the interfaces based on the discrete curl of an electric
field defined at cell edges.

NOTE: While the scheme theoretically keeps div B = 0,
floating point errors seem to accumulate over time,
especially in single precision. Projecting this divergence
out seemed to help with the divergence of the magnetic field
but comes at additional cost.
"""

import jax
from functools import partial
from typing import Tuple, Union
from jaxtyping import Array, Float

from astronomix._finite_difference._maths._differencing import finite_difference_int6
from astronomix._finite_difference._maths._interpolate import interp_center_to_face
from astronomix._finite_difference._maths._interpolate import interp_face_to_center
from astronomix._finite_difference._maths._interpolate import point_values_to_averages
from astronomix.variable_registry.registered_variables import RegisteredVariables

XAXIS = 0
YAXIS = 1
ZAXIS = 2

@partial(jax.jit, static_argnames=["registered_variables"])
def constrained_transport_rhs(
    conserved_state,
    weno_flux_x,
    weno_flux_y,
    weno_flux_z,
    dtdx, 
    dtdy,
    dtdz,
    registered_variables: RegisteredVariables,
):
    """
    Here we compute the RHS updates for the interface magnetic fields.
    """

    # Step 0: retrieve variables

    # at index i, the fluxes are at interfaces i+1/2
    By_flux_x_interface = weno_flux_x[registered_variables.magnetic_index.y]
    Bz_flux_x_interface = weno_flux_x[registered_variables.magnetic_index.z]
    Bx_flux_y_interface = weno_flux_y[registered_variables.magnetic_index.x]
    Bz_flux_y_interface = weno_flux_y[registered_variables.magnetic_index.z]
    Bx_flux_z_interface = weno_flux_z[registered_variables.magnetic_index.x]
    By_flux_z_interface = weno_flux_z[registered_variables.magnetic_index.y]

    # cell-centered variables
    rho = conserved_state[registered_variables.density_index]
    vx = conserved_state[registered_variables.momentum_index.x] / rho
    vy = conserved_state[registered_variables.momentum_index.y] / rho
    vz = conserved_state[registered_variables.momentum_index.z] / rho
    Bx = conserved_state[registered_variables.magnetic_index.x]
    By = conserved_state[registered_variables.magnetic_index.y]
    Bz = conserved_state[registered_variables.magnetic_index.z]

    # Step 1: Compute modified magnetic field fluxes (Eq. 12 - 17)

    # At x-interfaces, after the interpolation at index i,
    # we have values at i+1/2
    Bx_x_interface = interp_center_to_face(Bx, XAXIS)
    vy_x_interface = interp_center_to_face(vy, XAXIS)
    vz_x_interface = interp_center_to_face(vz, XAXIS)

    By_flux_x_interface_mod = By_flux_x_interface + Bx_x_interface * vy_x_interface
    Bz_flux_x_interface_mod = Bz_flux_x_interface + Bx_x_interface * vz_x_interface

    # At y-interfaces
    By_y_interface = interp_center_to_face(By, YAXIS)
    vx_y_interface = interp_center_to_face(vx, YAXIS)
    vz_y_interface = interp_center_to_face(vz, YAXIS)

    Bx_flux_y_interface_mod = Bx_flux_y_interface + By_y_interface * vx_y_interface
    Bz_flux_y_interface_mod = Bz_flux_y_interface + By_y_interface * vz_y_interface

    # At z-interfaces
    Bz_z_interface = interp_center_to_face(Bz, ZAXIS)
    vx_z_interface = interp_center_to_face(vx, ZAXIS)
    vy_z_interface = interp_center_to_face(vy, ZAXIS)

    Bx_flux_z_interface_mod = Bx_flux_z_interface + Bz_z_interface * vx_z_interface
    By_flux_z_interface_mod = By_flux_z_interface + Bz_z_interface * vy_z_interface

    # Step 2: Compute electric field components at cell edges (Equations 19-21)

    # interpolate from the y interfaces to the (x,y) edges
    g_star_x_edge = interp_center_to_face(Bx_flux_y_interface_mod, XAXIS)

    # interpolate from the x interfaces to the (x,y) edges
    f_star_y_edge = interp_center_to_face(By_flux_x_interface_mod, YAXIS)

    # electric field component at (x,y) edges
    Omega_z_edge = g_star_x_edge - f_star_y_edge

    # interpolate from the z interfaces to the (y,z) edges
    h_star_y_edge = interp_center_to_face(By_flux_z_interface_mod, YAXIS)

    # interpolate from the y interfaces to the (y,z) edges
    g_star_z_edge = interp_center_to_face(Bz_flux_y_interface_mod, ZAXIS)

    # electric field component at (y,z) edges
    Omega_x_edge = h_star_y_edge - g_star_z_edge

    # interpolate from the x interfaces to the (z,x) edges
    f_star_z_edge = interp_center_to_face(Bz_flux_x_interface_mod, ZAXIS)

    # interpolate from the z interfaces to the (z,x) edges
    h_star_x_edge = interp_center_to_face(Bx_flux_z_interface_mod, XAXIS)

    # electric field component at (z,x) edges
    Omega_y_edge = f_star_z_edge - h_star_x_edge

    # Step 3: Maintain high-order accuracy in spite of
    # dimensional splitting by converting point values to averages
    # In the paper (Seo & Ryu 2023) they say "the advective
    # fluxes are modified to approximate “point values” at grid
    # cell edges", but the formula used (also the finite differencing)
    # is that from point values to averages, compare 
    # Buchmüller & Helzel 2014.
    Omega_z_bar = point_values_to_averages(Omega_z_edge, XAXIS, YAXIS)
    Omega_x_bar = point_values_to_averages(Omega_x_edge, YAXIS, ZAXIS)
    Omega_y_bar = point_values_to_averages(Omega_y_edge, XAXIS, ZAXIS)

    # Update the interface magnetic fields based on the discrete curl of
    # the electric field at edges (Eq. 24 - 26)
    rhs_bx = - dtdy * finite_difference_int6(Omega_z_bar, YAXIS) \
             + dtdz * finite_difference_int6(Omega_y_bar, ZAXIS)

    rhs_by = - dtdz * finite_difference_int6(Omega_x_bar, ZAXIS) \
             + dtdx * finite_difference_int6(Omega_z_bar, XAXIS)

    rhs_bz = - dtdx * finite_difference_int6(Omega_y_bar, XAXIS) \
             + dtdy * finite_difference_int6(Omega_x_bar, YAXIS)

    return rhs_bx, rhs_by, rhs_bz


# Finite difference derivatives
# Interpolation from interfaces back to centers
@partial(jax.jit, static_argnames=["registered_variables"])
def update_cell_center_fields(
    conserved_state,
    bx_interface,
    by_interface,
    bz_interface,
    registered_variables: RegisteredVariables,
):
    """
    Update cell-centered B field from interface values
    using 6th order interpolation. Update the total energy
    accordingly to conserve total energy.
    """

    BX = registered_variables.magnetic_index.x
    BY = registered_variables.magnetic_index.y
    BZ = registered_variables.magnetic_index.z

    b2_old = (
        conserved_state[BX] ** 2 + conserved_state[BY] ** 2 + conserved_state[BZ] ** 2
    )

    # interpolate from interfaces back to cell centers
    Bx_center = interp_face_to_center(bx_interface, XAXIS)
    By_center = interp_face_to_center(by_interface, YAXIS)
    Bz_center = interp_face_to_center(bz_interface, ZAXIS)

    conserved_new = conserved_state.at[BX].set(Bx_center)
    conserved_new = conserved_new.at[BY].set(By_center)
    conserved_new = conserved_new.at[BZ].set(Bz_center)

    b2_new = conserved_new[BX] ** 2 + conserved_new[BY] ** 2 + conserved_new[BZ] ** 2

    # update total energy: E_new = E_old + 0.5 * (b2_new - b2_old)
    conserved_new = conserved_new.at[registered_variables.pressure_index].add(
        0.5 * (b2_new - b2_old)
    )

    return conserved_new


@jax.jit
def initialize_interface_fields(
    magnetic_field_x,
    magnetic_field_y,
    magnetic_field_z,
):
    """Initialize magnetic field at interfaces from cell centers."""

    # Use fourth-order interpolation
    bx_interface = interp_center_to_face(magnetic_field_x, XAXIS)
    by_interface = interp_center_to_face(magnetic_field_y, YAXIS)
    bz_interface = interp_center_to_face(magnetic_field_z, ZAXIS)

    return bx_interface, by_interface, bz_interface


import jax
from functools import partial

from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker

from typing import Union

from astronomix.option_classes.simulation_config import CYLINDRICAL, SPHERICAL


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["geometry"])
def _r_hat_alpha(
    r: Float[Array, "num_cells"], dr: Union[float, Float[Array, ""]], geometry: int
) -> Float[Array, "num_cells"]:
    if geometry == SPHERICAL:
        return r**2 + 1 / 12 * dr**2
    elif geometry == CYLINDRICAL:
        return r
    else:
        raise ValueError("Unknown geometry / not for cartesian coordinates")


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["geometry"])
def _center_of_volume(
    r: Float[Array, "num_cells"], dr: Union[float, Float[Array, ""]], geometry: int
) -> Float[Array, "num_cells"]:
    if geometry == CYLINDRICAL:
        return (r**2 + 1 / 12 * dr**2) / r**2 * r
    elif geometry == SPHERICAL:
        r_hat = _r_hat_alpha(r, dr, geometry)
        return (r**2 + 1 / 4 * dr**2) / r_hat * r



"""
Strong Stability Preserving Runge-Kutta (SSPRK) time integrator.

See _magnetic_update/_constrained_transport.py for more details on the
Constrained Transport (CT) implementation following (Seo & Ryu 2023,
https://arxiv.org/abs/2304.04360).
"""

from functools import partial
import jax
import jax.numpy as jnp
from typing import Union, Tuple

from astronomix._finite_difference._fluid_equations._enforce_positivity import (
    _enforce_positivity,
)
from astronomix._finite_difference._interface_fluxes._weno import (
    _weno_flux_x,
    _weno_flux_y,
    _weno_flux_z,
)

from astronomix._finite_difference._magnetic_update._constrained_transport import (
    constrained_transport_rhs,
    update_cell_center_fields,
)
from astronomix._finite_difference._maths._differencing import finite_difference_int6
from astronomix._physics_modules._stellar_wind.stellar_wind import _wind_ei3D_source
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.option_classes.simulation_config import SimulationConfig
from astronomix.option_classes.simulation_params import SimulationParams
from astronomix.variable_registry.registered_variables import RegisteredVariables


# @partial(jax.jit, static_argnames=["registered_variables", "config"], donate_argnames=["conserved_state", "bx_interface", "by_interface", "bz_interface"])
# def _ssprk4_with_ct(
#     conserved_state,
#     bx_interface,
#     by_interface,
#     bz_interface,
#     gamma: Union[float, jnp.ndarray],
#     grid_spacing: Union[float, jnp.ndarray],
#     dt: Union[float, jnp.ndarray],
#     params: SimulationParams,
#     helper_data: HelperData,
#     config: SimulationConfig,
#     registered_variables: RegisteredVariables,
# ):
#     """
#     Integrates the MHD equations for one time step using a 5-stage, 4th-order
#     Strong Stability Preserving Runge-Kutta (SSPRK) method
#     with Constrained Transport (CT).
#     """

#     def compute_rhs(current_q, bx, by, bz, k2_coeff):
#         """
#         Computes the right-hand side (RHS) of the MHD equations for a given stage.
#         The `k2_coeff` scales the timestep `dt` for the current RK stage.
#         """

#         current_q = update_cell_center_fields(
#             current_q, bx, by, bz, registered_variables
#         )

#         # in the future we might support
#         # different grid spacings in each direction
#         dtdx = k2_coeff * dt / grid_spacing
#         dtdy = k2_coeff * dt / grid_spacing
#         dtdz = k2_coeff * dt / grid_spacing

#         # Calculate fluxes based on the state of the current stage
#         dF_x = _weno_flux_x(current_q, params.minimum_density, params.minimum_pressure, gamma, registered_variables)
#         dF_y = _weno_flux_y(current_q, params.minimum_density, params.minimum_pressure, gamma, registered_variables)
#         dF_z = _weno_flux_z(current_q, params.minimum_density, params.minimum_pressure, gamma, registered_variables)

#         # Calculate RHS for interface magnetic fields using Constrained Transport
#         rhs_bx, rhs_by, rhs_bz = constrained_transport_rhs(
#             current_q, dF_x, dF_y, dF_z, dtdx, dtdy, dtdz, registered_variables
#         )

#         # Calculate RHS for conserved fluid variables
#         rhs_q = -dtdx * (
#             (dF_x - jnp.roll(dF_x, 1, axis=1))
#             + (dF_y - jnp.roll(dF_y, 1, axis=2))
#             + (dF_z - jnp.roll(dF_z, 1, axis=3))
#         )

#         if config.wind_config.stellar_wind:
#             rhs_q += (
#                 _wind_ei3D_source(
#                     params.wind_params,
#                     current_q,
#                     config,
#                     helper_data,
#                     config.wind_config.num_injection_cells,
#                     registered_variables,
#                 )
#                 * k2_coeff
#                 * dt
#             )

#         return rhs_q, rhs_bx, rhs_by, rhs_bz

#     # Store the initial state (t = n)
#     q0 = conserved_state
#     bx0, by0, bz0 = bx_interface, by_interface, bz_interface

#     # Stage 1
#     k1_1 = 1.0
#     k2_1 = 0.39175222700392
#     # k3_1 = 0.0, so it's omitted

#     rhs_q0, rhs_bx0, rhs_by0, rhs_bz0 = compute_rhs(q0, bx0, by0, bz0, k2_1)

#     q1 = q0 + rhs_q0
#     bx1, by1, bz1 = bx0 + rhs_bx0, by0 + rhs_by0, bz0 + rhs_bz0

#     # Stage 2
#     k1_2 = 0.44437049406734
#     k2_2 = 0.36841059262959
#     k3_2 = 0.55562950593266

#     rhs_q1, rhs_bx1, rhs_by1, rhs_bz1 = compute_rhs(q1, bx1, by1, bz1, k2_2)

#     q2 = k1_2 * q0 + k3_2 * q1 + rhs_q1
#     bx2 = k1_2 * bx0 + k3_2 * bx1 + rhs_bx1
#     by2 = k1_2 * by0 + k3_2 * by1 + rhs_by1
#     bz2 = k1_2 * bz0 + k3_2 * bz1 + rhs_bz1

#     # Stage 3
#     k1_3 = 0.62010185138540
#     k2_3 = 0.25189177424738
#     k3_3 = 0.37989814861460

#     rhs_q2, rhs_bx2, rhs_by2, rhs_bz2 = compute_rhs(q2, bx2, by2, bz2, k2_3)

#     q3 = k1_3 * q0 + k3_3 * q2 + rhs_q2
#     bx3 = k1_3 * bx0 + k3_3 * bx2 + rhs_bx2
#     by3 = k1_3 * by0 + k3_3 * by2 + rhs_by2
#     bz3 = k1_3 * bz0 + k3_3 * bz2 + rhs_bz2

#     # Stage 4
#     k1_4 = 0.17807995410773
#     k2_4 = 0.54497475021237
#     k3_4 = 0.82192004589227

#     rhs_q3, rhs_bx3, rhs_by3, rhs_bz3 = compute_rhs(q3, bx3, by3, bz3, k2_4)

#     q4 = k1_4 * q0 + k3_4 * q3 + rhs_q3
#     bx4 = k1_4 * bx0 + k3_4 * bx3 + rhs_bx3
#     by4 = k1_4 * by0 + k3_4 * by3 + rhs_by3
#     bz4 = k1_4 * bz0 + k3_4 * bz3 + rhs_bz3

#     # Stage 5 (Final Stage)
#     k1_5 = -2.081261929715610e-02
#     k2_5 = 0.22600748319395
#     k3_5 = 5.03580947213895e-01  # This corresponds to k3 in the Fortran code
#     k4_5 = 0.51723167208978  # This corresponds to k4
#     k5_5 = -6.518979800418380e-12  # This corresponds to k5

#     rhs_q4, rhs_bx4, rhs_by4, rhs_bz4 = compute_rhs(q4, bx4, by4, bz4, k2_5)

#     q5 = (k1_5 * q0) + (k4_5 * q2) + (k5_5 * q3) + (k3_5 * q4) + rhs_q4
#     bx_final = (k1_5 * bx0) + (k4_5 * bx2) + (k5_5 * bx3) + (k3_5 * bx4) + rhs_bx4
#     by_final = (k1_5 * by0) + (k4_5 * by2) + (k5_5 * by3) + (k3_5 * by4) + rhs_by4
#     bz_final = (k1_5 * bz0) + (k4_5 * bz2) + (k5_5 * bz3) + (k3_5 * bz4) + rhs_bz4

#     # Update the cell-centered magnetic fields in the conserved state array
#     # from the final interface magnetic fields.
#     q_final = update_cell_center_fields(
#         q5, bx_final, by_final, bz_final, registered_variables
#     )

#     return q_final, bx_final, by_final, bz_final


@partial(jax.jit, static_argnames=["registered_variables", "config"], donate_argnames=["conserved_state", "bx_interface", "by_interface", "bz_interface"])
def _ssprk4_with_ct(
    conserved_state,
    bx_interface,
    by_interface,
    bz_interface,
    gamma: Union[float, jnp.ndarray],
    grid_spacing: Union[float, jnp.ndarray],
    dt: Union[float, jnp.ndarray],
    params: SimulationParams,
    helper_data: HelperData,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
):
    """
    Integrates the MHD equations for one time step using a 5-stage, 4th-order
    Strong Stability Preserving Runge-Kutta (SSPRK) method
    with Constrained Transport (CT).
    """

    def compute_rhs(current_q, bx, by, bz, k2_coeff):
        """
        Computes the right-hand side (RHS) of the MHD equations for a given stage.
        The `k2_coeff` scales the timestep `dt` for the current RK stage.
        """

        current_q = update_cell_center_fields(
            current_q, bx, by, bz, registered_variables
        )

        # in the future we might support
        # different grid spacings in each direction
        dtdx = k2_coeff * dt / grid_spacing
        dtdy = k2_coeff * dt / grid_spacing
        dtdz = k2_coeff * dt / grid_spacing

        # Calculate fluxes based on the state of the current stage
        dF_x = _weno_flux_x(current_q, params.minimum_density, params.minimum_pressure, gamma, config, registered_variables)
        dF_y = _weno_flux_y(current_q, params.minimum_density, params.minimum_pressure, gamma, config, registered_variables)
        dF_z = _weno_flux_z(current_q, params.minimum_density, params.minimum_pressure, gamma, config, registered_variables)

        # Calculate RHS for interface magnetic fields using Constrained Transport
        rhs_bx, rhs_by, rhs_bz = constrained_transport_rhs(
            current_q, dF_x, dF_y, dF_z, dtdx, dtdy, dtdz, registered_variables
        )

        # Calculate RHS for conserved fluid variables
        rhs_q = -dtdx * (
            (dF_x - jnp.roll(dF_x, 1, axis=1))
            + (dF_y - jnp.roll(dF_y, 1, axis=2))
            + (dF_z - jnp.roll(dF_z, 1, axis=3))
        )

        if config.wind_config.stellar_wind:
            rhs_q += (
                _wind_ei3D_source(
                    params.wind_params,
                    current_q,
                    k2_coeff * dt,
                    config,
                    helper_data,
                    config.wind_config.num_injection_cells,
                    registered_variables,
                )
            )

        return rhs_q, rhs_bx, rhs_by, rhs_bz

    # define the SSPRK4 coefficients

    k1_1 = 1.0
    k2_1 = 0.39175222700392
    k3_1 = 0.0

    k1_2 = 0.44437049406734
    k2_2 = 0.36841059262959
    k3_2 = 0.55562950593266

    k1_3 = 0.62010185138540
    k2_3 = 0.25189177424738
    k3_3 = 0.37989814861460
    
    k1_4 = 0.17807995410773
    k2_4 = 0.54497475021237
    k3_4 = 0.82192004589227

    k1_5 = -2.081261929715610e-02
    k2_5 = 0.22600748319395
    k3_5 = 5.03580947213895e-01  # This corresponds to k3 in the Fortran code
    k4_5 = 0.51723167208978  # This corresponds to k4
    k5_5 = -6.518979800418380e-12  # This corresponds to k5

    final_factors = jnp.array([k1_5, 0.0, k4_5, k5_5, k3_5])
    k_rhs_s = jnp.array([k2_1, k2_2, k2_3, k2_4, k2_5])
    k_0_s = jnp.array([k1_1, k1_2, k1_3, k1_4, k1_5])
    k_curr_s = jnp.array([k3_1, k3_2, k3_3, k3_4, k3_5])

    # Store the initial state (t = n)
    q0 = conserved_state
    bx0, by0, bz0 = bx_interface, by_interface, bz_interface

    def ssprk_stage(stage_idx, carry):

        # unpack carry
        current_state, final_state = carry
        q_curr, bx_curr, by_curr, bz_curr = current_state
        q_final, bx_final, by_final, bz_final = final_state

        if config.enforce_positivity:
            q_curr = _enforce_positivity(
                q_curr,
                gamma,
                params.minimum_density,
                params.minimum_pressure,
                registered_variables,
            )

        k_rhs = k_rhs_s[stage_idx]
        k_0 = k_0_s[stage_idx]
        k_curr = k_curr_s[stage_idx]

        # update the current state
        rhs_q, rhs_bx, rhs_by, rhs_bz = compute_rhs(q_curr, bx_curr, by_curr, bz_curr, k_rhs)
        q_curr = k_0 * q0 + k_curr * q_curr + rhs_q
        bx_curr = k_0 * bx0 + k_curr * bx_curr + rhs_bx
        by_curr = k_0 * by0 + k_curr * by_curr + rhs_by
        bz_curr = k_0 * bz0 + k_curr * bz_curr + rhs_bz

        # update the final state
        final_factor = final_factors[stage_idx + 1]
        q_final += q_curr * final_factor
        bx_final += bx_curr * final_factor
        by_final += by_curr * final_factor
        bz_final += bz_curr * final_factor

        return (
            (q_curr, bx_curr, by_curr, bz_curr), 
            (q_final, bx_final, by_final, bz_final)
        )

    (
        (q4, bx4, by4, bz4),
        (q_final, bx_final, by_final, bz_final)
    ) = jax.lax.fori_loop(0, 4, ssprk_stage, 
        (
            (q0, bx0, by0, bz0), 
            (final_factors[0] * q0, final_factors[0] * bx0, final_factors[0] * by0, final_factors[0] * bz0)
        )
    )

    # Final Stage (Stage 5)
    rhs_q4, rhs_bx4, rhs_by4, rhs_bz4 = compute_rhs(q4, bx4, by4, bz4, k2_5)
    q_final = q_final + rhs_q4
    bx_final = bx_final + rhs_bx4
    by_final = by_final + rhs_by4
    bz_final = bz_final + rhs_bz4

    # Update the cell-centered magnetic fields in the conserved state array
    # from the final interface magnetic fields.
    q_final = update_cell_center_fields(
        q_final, bx_final, by_final, bz_final, registered_variables
    )

    if config.enforce_positivity:
        q_final = _enforce_positivity(
            q_final,
            gamma,
            params.minimum_density,
            params.minimum_pressure,
            registered_variables,
        )

    return q_final, bx_final, by_final, bz_final